(*
** Copyright (c) 1995, 3Dfx Interactive, Inc.
** All Rights Reserved.
**
** This is UNPUBLISHED PROPRIETARY SOURCE CODE of 3Dfx Interactive, Inc.;
** the contents of this file may not be disclosed to third parties, copied or
** duplicated in any form, in whole or in part, without the prior written
** permission of 3Dfx Interactive, Inc.
**
** RESTRICTED RIGHTS LEGEND:
** Use, duplication or disclosure by the Government is subject to restrictions
** as set forth in subdivision (c)(1)(ii) of the Rights in Technical Data
** and Computer Software clause at DFARS 252.227-7013, and/or in similar or
** successor clauses in the FAR, DOD or NASA FAR Supplement. Unpublished -
** rights reserved under the Copyright Laws of the United States.
*)

(***************************************************************************
 *
 *  Glide 2.43 Delphi adaptation by Erik Unger and Eric Nowinski
 *
 *  The Headers 3Dfx, Sst1vid, GlideSys, GlideUtl and Glide
 *  are included in this Glide.pas unit.
 *
 *  31.7.98
 *
 *  Download: http://www.bigfoot.com/~ungerik/
 *  E-Mail:
 *         Erik Unger <ungerik@bigfoot.com>
 *         Eric Nowinski <Eric.Nowinski@capway.com>
 *
 ***************************************************************************)

unit Glide2x;

{$MINENUMSIZE 4}
{$ALIGN ON}

interface

{$IFNDEF VER110}
type
  LongWord = LongInt;
{$ENDIF}

// A global variable to say if the DLL is loaded.
// Check before use Glide functions, or use the older version.
var
  GlideUnitIsInitalised : Boolean;

{$DEFINE GLIDE2}

const
{$IFDEF GLIDE2}
  GlideDLL = 'Glide2x.dll';
{$ENDIF}

{$IFDEF GLIDE3}
  GlideDLL = 'Glide3x.dll';
{$ENDIF}

////////////////////////////////////////////////////////////////////////////////
// 3Dfx.h
////////////////////////////////////////////////////////////////////////////////

// basic data types
type
  TFxU8      = Byte;
  TFxI8      = Shortint;
  TFxU16     = Word;
  TFxI16     = Smallint;
  TFxU32     = LongWord;
  TFxI32     = LongInt;
  TFxBool    = LongBool;
  TFxFloat   = Single;
  TFxDouble  = Double;

// color types
type
  TFxColor = LongInt;
  TFxColor4  = packed record
    r,g,b,a : TFxFloat;
  end;

// fundamental types
const
  FXTRUE = True;
  FXFALSE = False;

(*
** helper macros
*)

function FXBIT(i: variant) : variant;
function FXUNUSED(a: variant) : boolean;

////////////////////////////////////////////////////////////////////////////////
// Sst1vid.h
////////////////////////////////////////////////////////////////////////////////

// Video defines
type
  TGrScreenRefresh = (
    GR_REFRESH_60Hz,
    GR_REFRESH_70Hz,
    GR_REFRESH_72Hz,
    GR_REFRESH_75Hz,
    GR_REFRESH_80Hz,
    GR_REFRESH_90Hz,
    GR_REFRESH_100Hz,
    GR_REFRESH_85Hz,
    GR_REFRESH_120Hz
  );
  
const
  GR_REFRESH_NONE = TGrScreenRefresh($ff);

type
  TGrScreenResolution = (
    GR_RESOLUTION_320x200,
    GR_RESOLUTION_320x240,
    GR_RESOLUTION_400x256,
    GR_RESOLUTION_512x384,
    GR_RESOLUTION_640x200,
    GR_RESOLUTION_640x350,
    GR_RESOLUTION_640x400,
    GR_RESOLUTION_640x480,
    GR_RESOLUTION_800x600,
    GR_RESOLUTION_960x720,
    GR_RESOLUTION_856x480,
    GR_RESOLUTION_512x256,
    GR_RESOLUTION_1024x768,
    GR_RESOLUTION_1280x1024,
    GR_RESOLUTION_1600x1200,
    GR_RESOLUTION_400x300
  );

const
  GR_RESOLUTION_NONE      = TGrScreenResolution($ff);
  GR_RESOLUTION_MIN       = GR_RESOLUTION_320x200;
  GR_RESOLUTION_MAX	  = GR_RESOLUTION_1600x1200;

////////////////////////////////////////////////////////////////////////////////
// GlideSys.h
////////////////////////////////////////////////////////////////////////////////

// Endianness is stored in bits [30:31]
const
  GLIDE_ENDIAN_SHIFT  = 30;
  GLIDE_ENDIAN_LITTLE = $1 shl GLIDE_ENDIAN_SHIFT;
  GLIDE_ENDIAN_BIG    = $2 shl GLIDE_ENDIAN_SHIFT;

// OS is stored in bits [0:6]

  GLIDE_OS_SHIFT   = $0;
  GLIDE_OS_UNIX    = $1;
  GLIDE_OS_DOS32   = $2;
  GLIDE_OS_WIN32   = $4;
  GLIDE_OS_SYSTEM7 = $8;
  GLIDE_OS_OS2     = $10;
  GLIDE_OS_OTHER   = $20; // For Proprietary Arcade HW

// Sim vs. Hardware is stored in bits [7:8]

  GLIDE_SST_SHIFT  = $7;
  GLIDE_SST_SIM    = $1 shl GLIDE_SST_SHIFT;
  GLIDE_SST_HW     = $2 shl GLIDE_SST_SHIFT;

// Hardware type is stored in bits [9:12]

  GLIDE_HW_SHIFT   = $9;
  GLIDE_HW_SST1    = $1 shl GLIDE_HW_SHIFT;
  GLIDE_HW_SST96   = $2 shl GLIDE_HW_SHIFT;
  GLIDE_HW_SSTH3   = $4 shl GLIDE_HW_SHIFT;
  GLIDE_HW_SST2    = $8 shl GLIDE_HW_SHIFT;


  GLIDE_ENDIAN = GLIDE_ENDIAN_LITTLE;
  GLIDE_OS = GLIDE_OS_WIN32;
  GLIDE_SST = GLIDE_SST_HW;
  GLIDE_HW = GLIDE_HW_SST1;

  GLIDE_PLATFORM = GLIDE_ENDIAN or GLIDE_OS or GLIDE_SST or GLIDE_HW;
  GLIDE_NUM_TMU  = 2;

////////////////////////////////////////////////////////////////////////////////
// Glide.h
////////////////////////////////////////////////////////////////////////////////

// -----------------------------------------------------------------------
// type DEFINITIONS
// -----------------------------------------------------------------------
type
  TGrColor = TFxU32;
  TGrAlpha = TFxU8;
  TGrMipMapId = TFxU32;
  TGrFog   = TFxU8;

// -----------------------------------------------------------------------
// CONSTANTS AND TYPES
// -----------------------------------------------------------------------
const
  MAX_NUM_SST            = 4;
  MAX_MIPMAPS_PER_SST    = 1024;
  GR_FOG_TABLE_SIZE      = 64;
  GR_NULL_MIPMAP_HANDLE  = -1;
  GR_ZDEPTHVALUE_NEAREST = $FFFF;
  GR_ZDEPTHVALUE_FARTHEST = $0000;
  GR_WDEPTHVALUE_NEAREST = $0000;
  GR_WDEPTHVALUE_FARTHEST = $FFFF;

  GR_MIPMAPLEVELMASK_EVEN = (1 shl 0);
  GR_MIPMAPLEVELMASK_ODD  = (1 shl 1);
  GR_MIPMAPLEVELMASK_BOTH = GR_MIPMAPLEVELMASK_EVEN or GR_MIPMAPLEVELMASK_ODD;

  GR_LODBIAS_BILINEAR  = 0.5;
  GR_LODBIAS_TRILINEAR = 0.0;

type
  TGrChipID = TFxI32;
const
  GR_TMU0 = $0;
  GR_TMU1 = $1;
  GR_TMU2 = $2;
  GR_FBI  = $3;

type
  TGrCombineFunction = (
    GR_COMBINE_FUNCTION_ZERO,
    GR_COMBINE_FUNCTION_LOCAL,
    GR_COMBINE_FUNCTION_LOCAL_ALPHA,
    GR_COMBINE_FUNCTION_SCALE_OTHER,
    GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL,
    GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA,
    GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL,
    GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL,
    GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL_ALPHA,
    GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL,
    GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL_ALPHA
  );

const
  GR_COMBINE_FUNCTION_NONE = GR_COMBINE_FUNCTION_ZERO;
  GR_COMBINE_FUNCTION_BLEND_OTHER = GR_COMBINE_FUNCTION_SCALE_OTHER;
  GR_COMBINE_FUNCTION_BLEND
      = GR_COMBINE_FUNCTION_SCALE_OTHER_MINUS_LOCAL_ADD_LOCAL;
  GR_COMBINE_FUNCTION_BLEND_LOCAL
      = GR_COMBINE_FUNCTION_SCALE_MINUS_LOCAL_ADD_LOCAL;

type
  TGrCombineFactor = (
    GR_COMBINE_FACTOR_ZERO,
    GR_COMBINE_FACTOR_LOCAL,
    GR_COMBINE_FACTOR_OTHER_ALPHA,
    GR_COMBINE_FACTOR_LOCAL_ALPHA,
    GR_COMBINE_FACTOR_TEXTURE_ALPHA,
    GR_COMBINE_FACTOR_LOD_FRACTION,
    GR_COMBINE_INVALID_6,
    GR_COMBINE_INVALID_7,
    GR_COMBINE_FACTOR_ONE,
    GR_COMBINE_FACTOR_ONE_MINUS_LOCAL,
    GR_COMBINE_FACTOR_ONE_MINUS_OTHER_ALPHA,
    GR_COMBINE_FACTOR_ONE_MINUS_LOCAL_ALPHA,
    GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA,
    GR_COMBINE_FACTOR_ONE_MINUS_LOD_FRACTION
  );
const
  GR_COMBINE_FACTOR_NONE = GR_COMBINE_FACTOR_ZERO;
  GR_COMBINE_FACTOR_DETAIL_FACTOR = GR_COMBINE_FACTOR_TEXTURE_ALPHA;
  GR_COMBINE_FACTOR_ONE_MINUS_DETAIL_FACTOR =
      GR_COMBINE_FACTOR_ONE_MINUS_TEXTURE_ALPHA;

type
  TGrCombineLocal = (
    GR_COMBINE_LOCAL_ITERATED,
    GR_COMBINE_LOCAL_CONSTANT,
    GR_COMBINE_LOCAL_DEPTH
  );
  
const
  GR_COMBINE_LOCAL_NONE = GR_COMBINE_LOCAL_CONSTANT;

type
  TGrCombineOther = (
    GR_COMBINE_OTHER_ITERATED,
    GR_COMBINE_OTHER_TEXTURE,
    GR_COMBINE_OTHER_CONSTANT
  );
const
  GR_COMBINE_OTHER_NONE = GR_COMBINE_OTHER_CONSTANT;

type
  TGrAlphaSource = (
    GR_ALPHASOURCE_CC_ALPHA,
    GR_ALPHASOURCE_ITERATED_ALPHA,
    GR_ALPHASOURCE_TEXTURE_ALPHA,
    GR_ALPHASOURCE_TEXTURE_ALPHA_TIMES_ITERATED_ALPHA
  );

  TGrColorCombineFnc = (
    GR_COLORCOMBINE_ZERO,
    GR_COLORCOMBINE_CCRGB,
    GR_COLORCOMBINE_ITRGB,
    GR_COLORCOMBINE_ITRGB_DELTA0,
    GR_COLORCOMBINE_DECAL_TEXTURE,
    GR_COLORCOMBINE_TEXTURE_TIMES_CCRGB,
    GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB,
    GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB_DELTA0,
    GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB_ADD_ALPHA,
    GR_COLORCOMBINE_TEXTURE_TIMES_ALPHA,
    GR_COLORCOMBINE_TEXTURE_TIMES_ALPHA_ADD_ITRGB,
    GR_COLORCOMBINE_TEXTURE_ADD_ITRGB,
    GR_COLORCOMBINE_TEXTURE_SUB_ITRGB,
    GR_COLORCOMBINE_CCRGB_BLEND_ITRGB_ON_TEXALPHA,
    GR_COLORCOMBINE_DIFF_SPEC_A,
    GR_COLORCOMBINE_DIFF_SPEC_B,
    GR_COLORCOMBINE_ONE
  );

type
  TGrAlphaBlendFnc = (
    GR_BLEND_ZERO,
    GR_BLEND_SRC_ALPHA,
    GR_BLEND_SRC_COLOR,
    GR_BLEND_DST_ALPHA,
    GR_BLEND_ONE,
    GR_BLEND_ONE_MINUS_SRC_ALPHA,
    GR_BLEND_ONE_MINUS_SRC_COLOR,
    GR_BLEND_ONE_MINUS_DST_ALPHA,
    GR_BLEND_RESERVED_8,
    GR_BLEND_RESERVED_9,
    GR_BLEND_RESERVED_A,
    GR_BLEND_RESERVED_B,
    GR_BLEND_RESERVED_C,
    GR_BLEND_RESERVED_D,
    GR_BLEND_RESERVED_E,
    GR_BLEND_ALPHA_SATURATE
  );

const
  GR_BLEND_DST_COLOR = GR_BLEND_SRC_COLOR;
  GR_BLEND_ONE_MINUS_DST_COLOR = GR_BLEND_ONE_MINUS_SRC_COLOR;
  GR_BLEND_PREFOG_COLOR = GR_BLEND_ALPHA_SATURATE;

type
  TGrAspectRatio = (
    GR_ASPECT_8x1,       // 8W x 1H
    GR_ASPECT_4x1,       // 4W x 1H
    GR_ASPECT_2x1,       // 2W x 1H
    GR_ASPECT_1x1,       // 1W x 1H
    GR_ASPECT_1x2,       // 1W x 2H
    GR_ASPECT_1x4,       // 1W x 4H
    GR_ASPECT_1x8        // 1W x 8H
  );

  TGrBuffer = (
    GR_BUFFER_FRONTBUFFER,
    GR_BUFFER_BACKBUFFER,
    GR_BUFFER_AUXBUFFER,
    GR_BUFFER_DEPTHBUFFER,
    GR_BUFFER_ALPHABUFFER,
    GR_BUFFER_TRIPLEBUFFER
  );

  TGrChromakeyMode = (
    GR_CHROMAKEY_DISABLE,
    GR_CHROMAKEY_ENABLE
  );

  TGrCmpFnc = (
    GR_CMP_NEVER,
    GR_CMP_LESS,
    GR_CMP_EQUAL,
    GR_CMP_LEQUAL,
    GR_CMP_GREATER,
    GR_CMP_NOTEQUAL,
    GR_CMP_GEQUAL,
    GR_CMP_ALWAYS
  );

  TGrColorFormat = (
    GR_COLORFORMAT_ARGB,
    GR_COLORFORMAT_ABGR,
    GR_COLORFORMAT_RGBA,
    GR_COLORFORMAT_BGRA
  );

  TGrCullMode = (
    GR_CULL_DISABLE,
    GR_CULL_NEGATIVE,
    GR_CULL_POSITIVE
  );

  TGrDepthBufferMode = (
    GR_DEPTHBUFFER_DISABLE,
    GR_DEPTHBUFFER_ZBUFFER,
    GR_DEPTHBUFFER_WBUFFER,
    GR_DEPTHBUFFER_ZBUFFER_COMPARE_TO_BIAS,
    GR_DEPTHBUFFER_WBUFFER_COMPARE_TO_BIAS
  );

  TGrDitherMode = (
    GR_DITHER_DISABLE,
    GR_DITHER_2x2,
    GR_DITHER_4x4
  );

type
  TGrFogMode = TFxI32;
const
  GR_FOG_DISABLE          = $0;
  GR_FOG_WITH_ITERATED_ALPHA = $1;
  GR_FOG_WITH_TABLE       = $2;
  GR_FOG_MULT2            = $100;
  GR_FOG_ADD2             = $200;

type
  TGrLock = TFxU32;
const
  GR_LFB_READ_ONLY  = $00;
  GR_LFB_WRITE_ONLY = $01;
  GR_LFB_IDLE       = $00;
  GR_LFB_NOIDLE     = $10;

type
  TGrLfbBypassMode = (
    GR_LFBBYPASS_DISABLE,
    GR_LFBBYPASS_ENABLE
  );

  TGrLfbWriteMode = (
    GR_LFBWRITEMODE_565, // RGB:RGB
    GR_LFBWRITEMODE_555, // RGB:RGB
    GR_LFBWRITEMODE_1555, // ARGB:ARGB
    GR_LFBWRITEMODE_RESERVED1,
    GR_LFBWRITEMODE_888, // RGB
    GR_LFBWRITEMODE_8888, // ARGB
    GR_LFBWRITEMODE_RESERVED2,
    GR_LFBWRITEMODE_RESERVED3,
    GR_LFBWRITEMODE_RESERVED4,
    GR_LFBWRITEMODE_RESERVED5,
    GR_LFBWRITEMODE_RESERVED6,
    GR_LFBWRITEMODE_RESERVED7,
    GR_LFBWRITEMODE_565_DEPTH, // RGB:DEPTH
    GR_LFBWRITEMODE_555_DEPTH, // RGB:DEPTH
    GR_LFBWRITEMODE_1555_DEPTH, // ARGB:DEPTH
    GR_LFBWRITEMODE_DEPTH_DEPTH // DEPTH:DEPTH
  );

const
  GR_LFBWRITEMODE_ZA16        = TGrLfbWriteMode($f); // DEPTH:DEPTH
  GR_LFBWRITEMODE_ALPHA_ALPHA = GR_LFBWRITEMODE_DEPTH_DEPTH;
  GR_LFBWRITEMODE_ANY       = TGrLfbWriteMode($FF);

type
  TGrOriginLocation = (
    GR_ORIGIN_UPPER_LEFT,
    GR_ORIGIN_LOWER_LEFT
  );
const
  GR_ORIGIN_ANY           = TGrOriginLocation($FF);

type
  TGrLfbInfo = packed record
    size : Integer;
    lfbPtr : Pointer;
    strideInBytes : TFxU32;
    writeMode : TGrLfbWriteMode;
    origin : TGrOriginLocation;
  end;

  TGrLOD = (
    GR_LOD_256,
    GR_LOD_128,
    GR_LOD_64,
    GR_LOD_32,
    GR_LOD_16,
    GR_LOD_8,
    GR_LOD_4,
    GR_LOD_2,
    GR_LOD_1
  );

  TGrMipMapMode = (
    GR_MIPMAP_DISABLE, // no mip mapping
    GR_MIPMAP_NEAREST, // use nearest mipmap
    GR_MIPMAP_NEAREST_DITHER // GR_MIPMAP_NEAREST + LOD dith
  );

  TGrSmoothingMode = (
    GR_SMOOTHING_DISABLE,
    GR_SMOOTHING_ENABLE
  );

  TGrTextureClampMode = (
    GR_TEXTURECLAMP_WRAP,
    GR_TEXTURECLAMP_CLAMP
  );

  TGrTextureCombineFnc = (
    GR_TEXTURECOMBINE_ZERO, // texout = 0 */
    GR_TEXTURECOMBINE_DECAL, // texout = texthis */
    GR_TEXTURECOMBINE_OTHER, // this TMU in passthru mode */
    GR_TEXTURECOMBINE_ADD, // tout = tthis + t(this+1) */
    GR_TEXTURECOMBINE_MULTIPLY, // texout = tthis * t(this+1) */
    GR_TEXTURECOMBINE_SUBTRACT, // Sutract from upstream TMU */
    GR_TEXTURECOMBINE_DETAIL, // detail--detail on tthis */
    GR_TEXTURECOMBINE_DETAIL_OTHER, // detail--detail on tthis+1 */
    GR_TEXTURECOMBINE_TRILINEAR_ODD, // trilinear--odd levels tthis*/
    GR_TEXTURECOMBINE_TRILINEAR_EVEN, //trilinear--even levels tthis*/
    GR_TEXTURECOMBINE_ONE // texout = $FFFFFFFF */
  );

  TGrTextureFilterMode = (
    GR_TEXTUREFILTER_POINT_SAMPLED,
    GR_TEXTUREFILTER_BILINEAR
  );

  TGrTextureFormat = (
    GR_TEXFMT_8BIT,
    GR_TEXFMT_YIQ_422,
    GR_TEXFMT_ALPHA_8, // (0..0xFF) alpha     */
    GR_TEXFMT_INTENSITY_8, // (0..0xFF) intensity */
    GR_TEXFMT_ALPHA_INTENSITY_44,
    GR_TEXFMT_P_8, // 8-bit palette */
    GR_TEXFMT_RSVD0,
    GR_TEXFMT_RSVD1,
    GR_TEXFMT_16BIT,
    GR_TEXFMT_AYIQ_8422,
    GR_TEXFMT_RGB_565,
    GR_TEXFMT_ARGB_1555,
    GR_TEXFMT_ARGB_4444,
    GR_TEXFMT_ALPHA_INTENSITY_88,
    GR_TEXFMT_AP_88, // 8-bit alpha 8-bit palette */
    GR_TEXFMT_RSVD2
  );

const
  GR_TEXFMT_RGB_332               = GR_TEXFMT_8BIT;
  GR_TEXFMT_ARGB_8332             = GR_TEXFMT_16BIT;

type
  TGrTexTable = (
    GR_TEXTABLE_NCC0,
    GR_TEXTABLE_NCC1,
    GR_TEXTABLE_PALETTE
  );

type
  TGrNCCTable = (
    GR_NCCTABLE_NCC0,
    GR_NCCTABLE_NCC1
  );

type
  TGrTexBaseRange = (
    GR_TEXBASE_32_TO_1,
    GR_TEXBASE_64,
    GR_TEXBASE_128,
    GR_TEXBASE_256
  );

const
{$IFDEF GLIDE3}
  GLIDE3_EXTRA_STATE = 68;
{$ELSE}
  GLIDE3_EXTRA_STATE = 0;
{$ENDIF}

  GLIDE_STATE_PAD_SIZE  = 312 + GLIDE3_EXTRA_STATE;
//#ifdef GLIDE_LIB
//typedef struct _TGrState_s TGrState;
//#else
//typedef struct _TGrState_s {
//  char pad[GLIDE_STATE_PAD_SIZE];
//} TGrState;
//#endif

type
  TGrState = packed record
    pad : array [0..GLIDE_STATE_PAD_SIZE-1] of char;
  end;

{$IFDEF GLIDE3}
const
(* Types of data in strips *)
  GR_FLOAT       = 0;
  GR_U8          = 1;

(* Parameters for strips *)
  GR_VERTEX      = 0;
  GR_COLOR       = 1;
  GR_TEXTURE0    = 2;
  GR_TEXTURE1    = 3;

(* Componenets for strips *)
(* vertex *)
  GR_VERTEX_XYZ  = 3;
  GR_VERTEX_XYZW = 4;
(* Color *)
  GR_COLOR_RGB   = 3;
  GR_COLOR_RGBA  = 4;
(* Texture *)
  GR_TEX_NONE    = 0;
  GR_TEX_ST      = 2;
  GR_TEX_STW     = 3;

(* Stuf for grDrawArray *)
  GR_POINTS              = 0;
  GR_LINE_STRIP          = 1;
  GR_POLYGON             = 2;
  GR_TRIANGLE_STRIP      = 3;
  GR_TRIANGLE_FAN        = 4;
  GR_TRIANGLES           = 5;

{$ENDIF}


// -----------------------------------------------------------------------
// STRUCTURES
// -----------------------------------------------------------------------

// 3DF texture file structs
type
  Gu3dfHeader = packed record
    width, height : TFxU32;
    small_lod, large_lod : Integer;
    aspect_ratio : TGrAspectRatio;
    format : TGrTextureFormat;
  end;

  GuNccTable = packed record
    yRGB : array [0..15] of TFxU8;
    iRGB : array [0..3, 0..2] of TFxI16;
    qRGB : array [0..3, 0..2] of TFxI16;
    packed_data : array [0..11] of TFxU32;
  end;

  GuTexPalette = packed record
    data : array [0..255] of TFxU32;
  end;

  GuTexTable = packed record
    case byte of
      0: (nccTable : GuNccTable);
      1: (palette  : GuTexPalette);
  end;

  Gu3dfInfo = packed record
    header : Gu3dfHeader;
    table : GuTexTable;
    data : Pointer;
    mem_required : TFxU32;    // memory required for mip map in bytes.
  end;

  TGrTexInfo = packed record
    smallLod : TGrLOD;
    largeLod : TGrLOD;
    aspectRatio : TGrAspectRatio;
    format : TGrTextureFormat;
    data : Pointer;
  end;

  TGrMipMapInfo = packed record
    sst : Integer;                    // SST where this texture map was stored  */
    valid : TFxBool;                  // set when this table entry is allocated*/
    width, height : Integer;
    aspect_ratio : TGrAspectRatio;         // aspect ratio of the mip map.  */
    data : Pointer;                  // actual texture data  */

    format : TGrTextureFormat;                    // format of the texture table */
    mipmap_mode : TGrMipMapMode;               // mip map mode for this texture */
    magfilter_mode : TGrTextureFilterMode;       // filtering to be used when magnified */
    minfilter_mode : TGrTextureFilterMode;       // filtering to be used with minified  */
    s_clamp_mode : TGrTextureClampMode;         // how this texture should be clamped in s */
    t_clamp_mode : TGrTextureClampMode;         // how this texture should be clamped in t */
    tLOD : TFxU32;                   // Register value for tLOD register */
    tTextureMode : TFxU32;           // Register value for tTextureMode not including non-texture specific bits */
    lod_bias : TFxU32;               // LOD bias of the mip map in preshifted 4.2*/
    lod_min, lod_max : TGrLOD;       // largest and smallest levels of detail  */
    tmu : Integer;                    // tmu on which this texture resides */
    odd_even_mask : TFxU32;          // mask specifying levels on this tmu  */
    tmu_base_address : TFxU32;       // base addr (in TMU mem) of this texture */
    trilinear : TFxBool;              // should we blend by lod? */

    nccable : GuNccTable;              // NCC compression table (optional) */
  end;

type
  TGrSstType = (
    GR_SSTTYPE_VOODOO,
    GR_SSTTYPE_SST96,
    GR_SSTTYPE_AT3D,
    GR_SSTTYPE_Voodoo2
  );

type
  TGrTMUConfig = packed record
    tmuRev : Integer;                // Rev of Texelfx chip */
    tmuRam : Integer;                // 1, 2, or 4 MB */
  end;

  TGrVoodooConfig = packed record
    fbRam : Integer;                         // 1, 2, or 4 MB
    fbiRev : Integer;                        // Rev of Pixelfx chip
    nTexelfx : Integer;                      // How many texelFX chips are there?
    sliDetect : TFxBool;                     // Is it a scan-line interleaved board?
    tmuConfig : array [0..GLIDE_NUM_TMU] of TGrTMUConfig;   // Configuration of the Texelfx chips */
  end;

  TGrSst96Config = packed record
    fbRam : Integer;                  // How much? */
    nTexelfx : Integer;
    tmuConfig : TGrTMUConfig;
  end;

  TGrVoodoo2Config = TGrVoodooConfig;
  
  TGrAT3DConfig = packed record
    rev : Integer;
  end;

  TsstBoard = packed record
    Ssttype : TGrSstType;             // Which hardware is it? */
    case TGrSstType of
      GR_SSTTYPE_VOODOO: (VoodooConfig : TGrVoodooConfig);
      GR_SSTTYPE_SST96: (SST96Config : TGrSst96Config);
      GR_SSTTYPE_AT3D: (AT3DConfig : TGrAT3DConfig);
      GR_SSTTYPE_Voodoo2: (Voodoo2Config : TGrVoodoo2Config);
  end;

  TGrHwConfiguration = packed record
    num_sst : Integer;
    SSTs : array [0..MAX_NUM_SST-1] of TsstBoard;
  end;

  TGrSstPerfStats = packed record
    pixelsIn : TFxU32;              // # pixels processed (minus buffer clears) */
    chromaFail : TFxU32;            // # pixels not drawn due to chroma key */
    zFuncFail : TFxU32;             // # pixels not drawn due to Z comparison */
    aFuncFail : TFxU32;             // # pixels not drawn due to alpha comparison */
    pixelsOut : TFxU32;             // # pixels drawn (including buffer clears) */
  end;

  TGrTmuVertex = packed record
    sow : Single;                   // s texture ordinate (s over w) */
    tow : Single;                   // t texture ordinate (t over w) */
    oow : Single;                   // 1/w (used mipmapping - really 0xfff/w) */
  end;


// TGrVertex
// If these are changed the C & assembly language trisetup routines MUST
// be changed, for they will no longer work.
  TGrVertex = packed record
    x, y, z : Single;                // X, Y, and Z of scrn space -- Z is ignored */
    r, g, b : Single;                // R, G, B, ([0..255.0]) */
    ooz : Single;                    // 65535/Z (used for Z-buffering) */
    a : Single;                      // Alpha [0..255.0] */
    oow : Single;                    // 1/W (used for W-buffering, texturing) */
    tmuvtx : array [0..1] of TGrTmuVertex; // eigentlich ja 0..2... aber dann geht's ja nicht!
  end;

// For indexing TGrVertex as a float *.
//   CHANGE THESE IF THE VERTEX STRUCTURE CHANGES!
const
  GR_VERTEX_X_OFFSET              = 0;
  GR_VERTEX_Y_OFFSET              = 1;
  GR_VERTEX_Z_OFFSET              = 2;
  GR_VERTEX_R_OFFSET              = 3;
  GR_VERTEX_G_OFFSET              = 4;
  GR_VERTEX_B_OFFSET              = 5;
  GR_VERTEX_OOZ_OFFSET            = 6;
  GR_VERTEX_A_OFFSET              = 7;
  GR_VERTEX_OOW_OFFSET            = 8;
  GR_VERTEX_SOW_TMU0_OFFSET       = 9;
  GR_VERTEX_TOW_TMU0_OFFSET       = 10;
  GR_VERTEX_OOW_TMU0_OFFSET       = 11;
  GR_VERTEX_SOW_TMU1_OFFSET       = 12;
  GR_VERTEX_TOW_TMU1_OFFSET       = 13;
  GR_VERTEX_OOW_TMU1_OFFSET       = 14;
  GR_VERTEX_SOW_TMU2_OFFSET       = 15;
  GR_VERTEX_TOW_TMU2_OFFSET       = 16;
  GR_VERTEX_OOW_TMU2_OFFSET       = 17;

type
  TGrLfbSrcFmt = TFxU32;
const
  GR_LFB_SRC_FMT_565          = $00;
  GR_LFB_SRC_FMT_555          = $01;
  GR_LFB_SRC_FMT_1555         = $02;
  GR_LFB_SRC_FMT_888          = $04;
  GR_LFB_SRC_FMT_8888         = $05;
  GR_LFB_SRC_FMT_565_DEPTH    = $0c;
  GR_LFB_SRC_FMT_555_DEPTH    = $0d;
  GR_LFB_SRC_FMT_1555_DEPTH   = $0e;
  GR_LFB_SRC_FMT_ZA16         = $0f;
  GR_LFB_SRC_FMT_RLE16        = $80;

type
  TGrPassthruMode = (
    GR_PASSTHRU_SHOW_VGA,
    GR_PASSTHRU_SHOW_SST1 
  );

type
  TGrHint = TFxU32;
const
  GR_HINTTYPE_MIN                 = 0;
  GR_HINT_STWHINT                 = 0;
  GR_HINT_FIFOCHECKHINT           = 1;
  GR_HINT_FPUPRECISION            = 2;
  GR_HINT_ALLOW_MIPMAP_DITHER     = 3;
  GR_HINTTYPE_MAX                 = 3;

type
  TGrSTWHint = TFxU32;
const
  GR_STWHINT_W_DIFF_FBI   = 1 shl 0;
  GR_STWHINT_W_DIFF_TMU0  = 1 shl 1;
  GR_STWHINT_ST_DIFF_TMU0 = 1 shl 2;
  GR_STWHINT_W_DIFF_TMU1  = 1 shl 3;
  GR_STWHINT_ST_DIFF_TMU1 = 1 shl 4;
  GR_STWHINT_W_DIFF_TMU2  = 1 shl 5;
  GR_STWHINT_ST_DIFF_TMU2 = 1 shl 6;

type
  TGrControl = (
    GR_CONTROL_INVALID_0,
    GR_CONTROL_ACTIVATE,
    GR_CONTROL_DEACTIVATE,
    GR_CONTROL_RESIZE,
    GR_CONTROL_MOVE
  );

function GR_GENERATE_FIFOCHECK_HINT_MASK(swHWM, swLWM: variant) : variant;

// -----------------------------------------------------------------------
// function PROTOTYPES
// -----------------------------------------------------------------------

// rendering functions
var
  grDrawLine : procedure (var v1, v2 : TGrVertex); stdcall;
  grDrawPlanarPolygon : procedure (nverts : Integer; var ilist : Integer; var vlist : TGrVertex); stdcall;
  grDrawPlanarPolygonVertexList : procedure (nverts : Integer; var vlist : TGrVertex); stdcall;
  grDrawPoint : procedure (var pt : TGrVertex); stdcall;
  grDrawPolygon : procedure (nverts : Integer; var ilist : Integer; var vlist : TGrVertex); stdcall;
  grDrawPolygonVertexList : procedure (nverts : Integer; var vlist : TGrVertex); stdcall;
  grDrawTriangle : procedure (var a, b, c : TGrVertex); stdcall;

// buffer management
var
  grBufferClear : procedure (Color : TGrColor; alpha : TGrAlpha; depth : TFxU16); stdcall;
  grBufferNumPending : procedure  ; stdcall;
  grBufferSwap : procedure (swap_interval : Integer); stdcall;
  grRenderBuffer : procedure (buffer : TGrBuffer); stdcall;

// error management
type
  TGrErrorCallbackFnc = procedure(str : PChar; fatal : TFxBool);
var
  grErrorSetCallback : procedure (fnc : TGrErrorCallbackFnc); stdcall;

// SST routines
var
  grSstIdle : procedure ; stdcall;
  grSstVideoLine : function : TFxU32; stdcall;
  grSstVRetraceOn : function : TFxBool; stdcall;
  grSstIsBusy : function : TFxBool; stdcall;
  grSstWinOpen : function (hWnd : TFxU32;
                           screen_resolution : TGrScreenResolution;
                           refresh_rate : TGrScreenRefresh;
                           color_format : TGrColorFormat;
                           origin_location : TGrOriginLocation;
                           nColBuffers, nAuxBuffers : Integer) : TFxBool; stdcall;
  grSstWinClose : procedure ; stdcall;
  grSstControl : function  (code : TFxU32): TFxBool; stdcall;
  grSstQueryHardware : function (var hwconfig : TGrHwConfiguration) : TFxBool; stdcall;
  grSstQueryBoards : function (var hwconfig : TGrHwConfiguration) : TFxBool; stdcall;
  grSstOrigin : procedure (origine : TGrOriginLocation); stdcall;
  grSstSelect : procedure (which_sst : Integer); stdcall;
  grSstScreenHeight : function : Integer; stdcall;
  grSstScreenWidth : function : Integer; stdcall;
  grSstStatus : function : TFxU32; stdcall;

//  Drawing Statistics
var
  grSstPerfStats : procedure (var pStats : TGrSstPerfStats); stdcall;
  grSstResetPerfStats : procedure ; stdcall;
  grResetTriStats : procedure ; stdcall;
  grTriStats : procedure (var trisProcessed : TFxU32; trisDrawn : TFxU32); stdcall;

// Glide configuration and special effect maintenance functions
var
  grAlphaBlendFunction : procedure (rgb_sf, rgb_df, alpha_sf, alpha_df : TGrAlphaBlendFnc); stdcall;
  grAlphaCombine : procedure (Fnct : TGrCombineFunction; factor : TGrCombineFactor;
                              local : TGrCombineLocal; other : TGrCombineOther;
                              invert : TFxBool); stdcall;
  grAlphaControlsITRGBLighting : procedure (enable : TFxBool); stdcall;
  grAlphaTestFunction : procedure (fnct : TGrCmpFnc); stdcall;
  grAlphaTestReferenceValue : procedure (value : TGrAlpha); stdcall;
  grChromakeyMode : procedure (mode : TGrChromakeyMode); stdcall;
  grChromakeyValue : procedure (value : TGrColor); stdcall;
  grClipWindow : procedure (minx, miny, maxx, maxy : Integer); stdcall;
  grColorCombine : procedure (Fnct : TGrCombineFunction; factor : TGrCombineFactor;
                              local : TGrCombineLocal; other :TGrCombineOther;
                              invert : TFxBool); stdcall;
  grColorMask : procedure (rgb, a : TFxBool); stdcall;
  grCullMode : procedure (mode : TGrCullMode); stdcall;
  grConstantColorValue : procedure (value : TGrColor); stdcall;
  grConstantColorValue4 : procedure (a, r, g, b : Single); stdcall;
  grDepthBiasLevel : procedure (level : TFxI16); stdcall;
  grDepthBufferFunction : procedure (fnct : TGrCmpFnc); stdcall;
  grDepthBufferMode : procedure (mode : TGrDepthBufferMode); stdcall;
  grDepthMask : procedure (mask : TFxBool); stdcall;
  grDisableAllEffects : procedure ; stdcall;
  grDitherMode : procedure (mode : TGrDitherMode); stdcall;
  grFogColorValue : procedure (fogcolor : TGrColor); stdcall;
  grFogMode : procedure (mode : TGrFogMode); stdcall;
  grFogTable : procedure (var ft : TGrFog); stdcall;
  grGammaCorrectionValue : procedure (value : single); stdcall;
  grSplash : procedure (x, y : Single;width, height : Single; frame : TFxU32); stdcall;

// texture mapping control functions
var
  grTexCalcMemRequired : function (lodmin, modmax : TGrLOD; aspect : TGrAspectRatio; fmt : TGrTextureFormat) : TFxU32; stdcall;
  grTexTextureMemRequired : function (evenOdd : TFxU32; var info : TGrTexInfo) : TFxU32; stdcall;
  grTexMinAddress : function (tmu : TGrChipID) : TFxU32; stdcall;
  grTexMaxAddress : function (tmu : TGrChipID) : TFxU32; stdcall;
  grTexNCCTable : procedure (tmu : TGrChipID ; table : TGrNCCTable); stdcall;
  grTexSource : procedure (tmu : TGrChipID; startAddess, evenOdd : TFxU32; var info : TGrTexInfo); stdcall;
  grTexClampMode : procedure (tmu : TGrChipID; s_clampmode, t_clampmode: TGrTextureClampMode); stdcall;
  grTexCombine : procedure (tmu : TGrChipID;
                            rgb_function : TGrCombineFunction;
                            rgb_factor : TGrCombineFactor;
                            alpha_function : TGrCombineFunction;
                            alpha_factor : TGrCombineFactor;
                            rgb_invert, alpha_invert : TFxBool); stdcall;
  grTexCombineFunction : procedure (tmu : TGrChipID; fnc : TGrTextureCombineFnc); stdcall;
  grTexDetailControl : procedure (tmu : TGrChipID; lod_bias : Integer; detail_scale : TFxU8; detail_max : Single); stdcall;
  grTexFilterMode : procedure (tmu : TGrChipID; minfilter_mode,magfilter_mode : TGrTextureFilterMode); stdcall;
  grTexLodBiasValue : procedure (tmu : TGrChipID; bias : Single); stdcall;
  grTexDownloadMipMap : procedure (tmu : TGrChipID; startAddress, enenOdd : TFxU32; var info : TGrTexInfo); stdcall;
  grTexDownloadMipMapLevel : procedure (tmu : TGrChipID;
                                        startAddress : TFxU32;
                                        thisLod, largeLod : TGrLOD;
                                        aspectRatio : TGrAspectRatio;
                                        format : TGrTextureFormat;
                                        evenOdd : TFxU32;
                                        data : Pointer); stdcall;
  grTexDownloadMipMapLevelPartial : procedure (tmu : TGrChipID;
                                               startAdress : TFxU32;
                                               thisLod, largeLod : TGrLOD;
                                               aspectRatio : TGrAspectRatio;
                                               format : TGrTextureFormat;
                                               evenOdd : TFxU32;
                                               data : Pointer;
                                               _start, _end : Integer); stdcall;
  ConvertAndDownloadRle : procedure (tmu : TGrChipID;
                                    startAddress : TFxU32;
                                    thisLod, largeLod : TGrLOD;
                                    aspectRatio : TGrAspectRatio;
                                    format : TGrTextureFormat;
                                    evenOdd : TFxU32;
                                    var bm_data : TFxU8;
                                    bm_h : LongInt;
                                    u0, v0 : TFxU32;
                                    width, height : TFxU32;
                                    dest_width, dest_height : TFxU32;
                                    var tlut : TFxU16); stdcall;
  grCheckForRoom : procedure (n : TFxI32); stdcall;
  grTexDownloadTable : procedure (tmu : TGrChipID; ype : TGrTexTable; data : Pointer); stdcall;
  grTexDownloadTablePartial : procedure (tmu : TGrChipID; _Type : TGrTexTable; data : Pointer;
                                         _start, _end : Integer); stdcall;
  grTexMipMapMode : procedure (tmu : TGrChipID; mode : TGrMipMapMode;lodBlend : TFxBool); stdcall;
  grTexMultibase : procedure (tmu : TGrChipID; enable : TFxBool); stdcall;
  grTexMultibaseAddress : procedure (tmu : TGrChipID; range : TGrTexBaseRange;
                                     startAddress, evenodd : TFxU32;
                                     var info : TGrTexInfo); stdcall;

// utility texture functions
var
  guTexAllocateMemory : function (tmu : TGrChipID;
                                  odd_even_mask : TFxU8;
                                  width, height : Integer;
                                  fmt : TGrTextureFormat;
                                  mm_mode : TGrMipMapMode;
                                  smallest_lod, largest_lod : TGrLOD;
                                  aspect : TGrAspectRatio;
                                  s_clamp_mode, t_clamp_mode : TGrTextureClampMode;
                                  minfilter_mode, magfilter_mode : TGrTextureFilterMode;
                                  lod_bias : Single;
                                  trilinear : TFxBool) : TGrMipMapId; stdcall;
  guTexChangeAttributes : function (mmid : TGrMipMapId;
                                    width, height : Integer;
                                    fmt : TGrTextureFormat;
                                    mm_mode : TGrMipMapMode;
                                    smallest_lod, largest_lod : TGrLOD;
                                    aspect : TGrAspectRatio;
                                    s_clamp_mode, t_clamp_mode : TGrTextureClampMode;
                                    minfilter_mode, magfilter_mode : TGrTextureFilterMode) : TFxBool; stdcall;
  guTexCombineFunction : procedure (tmu : TGrChipID; fnc : TGrTextureCombineFnc); stdcall;
  guTexGetCurrentMipMap : function (tmu : TGrChipID) : TGrMipMapId; stdcall;
  guTexGetMipMapInfo : function (mmid : TGrMipMapId) : TGrMipMapInfo; stdcall;
  guTexMemQueryAvail : function (tmu : TGrChipID) : TFxU32; stdcall;
  guTexMemReset : procedure ; stdcall;
  guTexDownloadMipMap : procedure (mmid : TGrMipMapId; src : Pointer; var table : GuNccTable); stdcall;
  guTexDownloadMipMapLevel : procedure (mmid : TGrMipMapId; lod : TGrLOD; var src : Pointer); stdcall;
  guTexSource : procedure (id : TGrMipMapId); stdcall;

// linear frame buffer functions
var
  grLfbLock : function (ype : TGrLock; buffer : TGrBuffer; writeMode : TGrLfbWriteMode;
                        origin : TGrOriginLocation; pixelPipeline : TFxBool;
                        var info : TGrLfbInfo) : TFxBool; stdcall;
  grLfbUnlock : function (ype : TGrLock; buffer : TGrBuffer) : TFxBool; stdcall;
  grLfbConstantAlpha : procedure (alpha : TGrAlpha); stdcall;
  grLfbConstantDepth : procedure (depth : TFxU16); stdcall;
  grLfbWriteColorSwizzle : procedure (swizzleBytes : TFxBool; swapWords : TFxBool); stdcall;
  grLfbWriteColorFormat : procedure (colorFormat : TGrColorFormat); stdcall;
  grLfbWriteRegion : function (dst_buffer : TGrBuffer;
                               dst_x, dst_y : TFxU32;
                               src_format : TGrLfbSrcFmt;
                               src_width, src_height : TFxU32;
                               src_stride : TFxI32;
                               src_data : Pointer) : TFxBool; stdcall;
  grLfbReadRegion : function (src_buffer : TGrBuffer;
                              src_x, src_y : TFxU32;
                              src_width, src_height : TFxU32;
                              dst_stride : TFxU32;
                              dst_data : Pointer) : TFxBool; stdcall;

//  Antialiasing Functions
var
  grAADrawLine : procedure (var v1 : TGrVertex; var v2 : TGrVertex); stdcall;
  grAADrawPoint : procedure (var pt : TGrVertex); stdcall;
  grAADrawPolygon : procedure (nverts : Integer; var ilist : Integer; var vlist : TGrVertex); stdcall;
  grAADrawPolygonVertexList : procedure (nverts : Integer; var vlist : TGrVertex); stdcall;
  grAADrawTriangle : procedure (var a, b, c : TGrVertex; ab_antialias, bc_antialias, ca_antialias : TFxBool); stdcall;

// glide management functions
var
  grGlideInit : procedure ; stdcall;
  grGlideShutdown : procedure ; stdcall;
  grGlideGetVersion : procedure (version : PChar); stdcall;
  grGlideGetState : procedure (var state : TGrState); stdcall;
  grGlideSetState : procedure (var state : TGrState); stdcall;
  grGlideShamelessPlug : procedure (_On : TFxBool); stdcall;
  grHints : procedure (hintType : TGrHint; hintMask : TFxU32); stdcall;

{$IFDEF GLIDE3}
var
  grParameterData : procedure (param,components,_type: TFxU32; offset: TFxI32); stdcall;
  grDrawArray : procedure (mode,Count: TFxU32; var pointers); stdcall;
{$ENDIF}

////////////////////////////////////////////////////////////////////////////////
// Glideutl.h
////////////////////////////////////////////////////////////////////////////////

// rendering functions
var
  guAADrawTriangleWithClip : procedure (var a, b, c: TGrVertex); stdcall;
  guDrawTriangleWithClip : procedure (var a, b, c: TGrVertex); stdcall;
  guDrawPolygonVertexListWithClip : procedure (nverts : Integer ;vlist : TGrVertex); stdcall;

// hi-level rendering utility functions
var
  guAlphaSource : procedure (mode : TGrAlphaSource); stdcall;
  guColorCombineFunction : procedure (fnc: TGrColorCombineFnc); stdcall;
  guEncodeRLE16 : function (dst, src: Pointer; width, height : TFxU32) : Integer; stdcall;
  guTexCreateColorMipMap : function  : Pointer; stdcall;

// fog stuff
var
  guFogTableIndexToW : function (i: Integer) : Single; stdcall;
  guFogGenerateExp : procedure (var fogtable: TGrFog; density: Single); stdcall;
  guFogGenerateExp2 : procedure (var fogtable: TGrFog; density: Single); stdcall;
  guFogGenerateLinear : procedure (var fogtable: TGrFog; nearZ, farZ: Single); stdcall;

// endian stuff
var
  guEndianSwapWords : function (value: TFxU32) : TFxU32; stdcall;
  guEndianSwapBytes : function (value: TFxU16) : TFxU16; stdcall;

// hi-level texture manipulation tools.
var
  gu3dfGetInfo : function (filename: PChar; var info: Gu3dfInfo): TFxBool; stdcall;
  gu3dfLoad : function (filename: PChar; var data: Gu3dfInfo): TFxBool; stdcall;

// movie capture stuff
var
  guMovieStart : procedure ; stdcall;
  guMovieStop : procedure ; stdcall;
  guMovieSetName : procedure (name: PChar); stdcall;

function InitGlideUnit : Boolean;
function CloseGlideUnit : Boolean;

implementation



uses
  Windows,
  SysUtils;


var
  GlideHandle : Integer;

function InitGlideUnit : Boolean;
begin
  result := False;
  if (GlideHandle = 0) then
  begin
    GlideHandle := LoadLibrary(GlideDLL);
    if (GlideHandle = 0) then
      exit;

    // rendering functions
    grDrawLine := GetProcAddress (GlideHandle, '_grDrawLine@8');
    grDrawPlanarPolygon := GetProcAddress (GlideHandle, '_grDrawPlanarPolygon@12');
    grDrawPlanarPolygonVertexList := GetProcAddress (GlideHandle, '_grDrawPlanarPolygonVertexList@8');
    grDrawPoint := GetProcAddress (GlideHandle, '_grDrawPoint@4');
    grDrawPolygon := GetProcAddress (GlideHandle, '_grDrawPolygon@12');
    grDrawPolygonVertexList := GetProcAddress (GlideHandle, '_grDrawPolygonVertexList@8');
    grDrawTriangle := GetProcAddress (GlideHandle, '_grDrawTriangle@12');

    // buffer management
    grBufferClear := GetProcAddress (GlideHandle, '_grBufferClear@12');
    grBufferNumPending := GetProcAddress (GlideHandle, '_grBufferNumPending@0');
    grBufferSwap := GetProcAddress (GlideHandle, '_grBufferSwap@4');
    grRenderBuffer := GetProcAddress (GlideHandle, '_grRenderBuffer@4');

    // error management
    grErrorSetCallback := GetProcAddress (GlideHandle, '_grErrorSetCallback@4');

    // SST routines
    grSstIdle := GetProcAddress (GlideHandle, '_grSstIdle@0');
    grSstVideoLine := GetProcAddress (GlideHandle, '_grSstVideoLine@0');
    grSstVRetraceOn := GetProcAddress (GlideHandle, '_grSstVRetraceOn@0');
    grSstIsBusy := GetProcAddress (GlideHandle, '_grSstIsBusy@0');
    grSstWinOpen := GetProcAddress (GlideHandle, '_grSstWinOpen@28');
    grSstWinClose := GetProcAddress (GlideHandle, '_grSstWinClose@0');
    grSstControl := GetProcAddress (GlideHandle, '_grSstControl@4');
    grSstQueryHardware := GetProcAddress (GlideHandle, '_grSstQueryHardware@4');
    grSstQueryBoards := GetProcAddress (GlideHandle, '_grSstQueryBoards@4');
    grSstOrigin := GetProcAddress (GlideHandle, '_grSstOrigin@4');
    grSstSelect := GetProcAddress (GlideHandle, '_grSstSelect@4');
    grSstScreenHeight := GetProcAddress (GlideHandle, '_grSstScreenHeight@0');
    grSstScreenWidth := GetProcAddress (GlideHandle, '_grSstScreenWidth@0');
    grSstStatus := GetProcAddress (GlideHandle, '_grSstStatus@0');
//    grSstPassthruMode := GetProcAddress (GlideHandle, '_grSstPassthruMode@4');

    //  Drawing Statistics
    grSstPerfStats := GetProcAddress (GlideHandle, '_grSstPerfStats@4');
    grSstResetPerfStats := GetProcAddress (GlideHandle, '_grSstResetPerfStats@0');
    grResetTriStats := GetProcAddress (GlideHandle, '_grResetTriStats@0');
    grTriStats := GetProcAddress (GlideHandle, '_grTriStats@8');

    // Glide configuration and special effect maintenance functions
    grAlphaBlendFunction := GetProcAddress (GlideHandle, '_grAlphaBlendFunction@16');
    grAlphaCombine := GetProcAddress (GlideHandle, '_grAlphaCombine@20');
    grAlphaControlsITRGBLighting := GetProcAddress (GlideHandle, '_grAlphaControlsITRGBLighting@4');
    grAlphaTestFunction := GetProcAddress (GlideHandle, '_grAlphaTestFunction@4');
    grAlphaTestReferenceValue := GetProcAddress (GlideHandle, '_grAlphaTestReferenceValue@4');
    grChromakeyMode := GetProcAddress (GlideHandle, '_grChromakeyMode@4');
    grChromakeyValue := GetProcAddress (GlideHandle, '_grChromakeyValue@4');
    grClipWindow := GetProcAddress (GlideHandle, '_grClipWindow@16');
    grColorCombine := GetProcAddress (GlideHandle, '_grColorCombine@20');
    grColorMask := GetProcAddress (GlideHandle, '_grColorMask@8');
    grCullMode := GetProcAddress (GlideHandle, '_grCullMode@4');
    grConstantColorValue := GetProcAddress (GlideHandle, '_grConstantColorValue@4');
    grConstantColorValue4 := GetProcAddress (GlideHandle, '_grConstantColorValue4@16');
    grDepthBiasLevel := GetProcAddress (GlideHandle, '_grDepthBiasLevel@4');
    grDepthBufferFunction := GetProcAddress (GlideHandle, '_grDepthBufferFunction@4');
    grDepthBufferMode := GetProcAddress (GlideHandle, '_grDepthBufferMode@4');
    grDepthMask := GetProcAddress (GlideHandle, '_grDepthMask@4');
    grDisableAllEffects := GetProcAddress (GlideHandle, '_grDisableAllEffects@0');
    grDitherMode := GetProcAddress (GlideHandle, '_grDitherMode@4');
    grFogColorValue := GetProcAddress (GlideHandle, '_grFogColorValue@4');
    grFogMode := GetProcAddress (GlideHandle, '_grFogMode@4');
    grFogTable := GetProcAddress (GlideHandle, '_grFogTable@4');
    grGammaCorrectionValue := GetProcAddress (GlideHandle, '_grGammaCorrectionValue@4');
    grSplash := GetProcAddress (GlideHandle, '_grSplash@20');

    // texture mapping control functions
    grTexCalcMemRequired := GetProcAddress (GlideHandle, '_grTexCalcMemRequired@16');
    grTexTextureMemRequired := GetProcAddress (GlideHandle, '_grTexTextureMemRequired@8');
    grTexMinAddress := GetProcAddress (GlideHandle, '_grTexMinAddress@4');
    grTexMaxAddress := GetProcAddress (GlideHandle, '_grTexMaxAddress@4');
    grTexNCCTable := GetProcAddress (GlideHandle, '_grTexNCCTable@8');
    grTexSource := GetProcAddress (GlideHandle, '_grTexSource@16');
    grTexClampMode := GetProcAddress (GlideHandle, '_grTexClampMode@12');
    grTexCombine := GetProcAddress (GlideHandle, '_grTexCombine@28');
    grTexCombineFunction := GetProcAddress (GlideHandle, '_grTexCombineFunction@8');
    grTexDetailControl := GetProcAddress (GlideHandle, '_grTexDetailControl@16');
    grTexFilterMode := GetProcAddress (GlideHandle, '_grTexFilterMode@12');
    grTexLodBiasValue := GetProcAddress (GlideHandle, '_grTexLodBiasValue@8');
    grTexDownloadMipMap := GetProcAddress (GlideHandle, '_grTexDownloadMipMap@16');
    grTexDownloadMipMapLevel := GetProcAddress (GlideHandle, '_grTexDownloadMipMapLevel@32');
    grTexDownloadMipMapLevelPartial := GetProcAddress (GlideHandle, '_grTexDownloadMipMapLevelPartial@40');
    grTexDownloadTable := GetProcAddress (GlideHandle, '_grTexDownloadTable@12');
    grTexDownloadTablePartial := GetProcAddress (GlideHandle, '_grTexDownloadTablePartial@20');
    ConvertAndDownloadRle := GetProcAddress (GlideHandle, '_ConvertAndDownloadRle@64');
    grCheckForRoom := GetProcAddress (GlideHandle, '_grCheckForRoom@4');
    grTexMipMapMode := GetProcAddress (GlideHandle, '_grTexMipMapMode@12');
    grTexMultibase := GetProcAddress (GlideHandle, '_grTexMultibase@8');
    grTexMultibaseAddress := GetProcAddress (GlideHandle, '_grTexMultibaseAddress@20');

    // utility texture functions
    guTexAllocateMemory := GetProcAddress (GlideHandle,  '_guTexAllocateMemory@60');
    guTexChangeAttributes := GetProcAddress (GlideHandle,  '_guTexChangeAttributes@48');
    guTexCombineFunction := GetProcAddress (GlideHandle,  '_guTexCombineFunction@8');
    guTexGetCurrentMipMap := GetProcAddress (GlideHandle,  '_guTexGetCurrentMipMap@4');
    guTexGetMipMapInfo := GetProcAddress (GlideHandle,  '_guTexGetMipMapInfo@4');
    guTexMemQueryAvail := GetProcAddress (GlideHandle,  '_guTexMemQueryAvail@4');
    guTexMemReset := GetProcAddress (GlideHandle,  '_guTexMemReset@0');
    guTexDownloadMipMap := GetProcAddress (GlideHandle,  '_guTexDownloadMipMap@12');
    guTexDownloadMipMapLevel := GetProcAddress (GlideHandle,  '_guTexDownloadMipMapLevel@12');
    guTexSource := GetProcAddress (GlideHandle,  '_guTexSource@4');

    // linear frame buffer functions
    grLfbLock := GetProcAddress (GlideHandle,  '_grLfbLock@24');
    grLfbUnLock := GetProcAddress (GlideHandle,  '_grLfbUnlock@8');
    grLfbConstantAlpha := GetProcAddress (GlideHandle,  '_grLfbConstantAlpha@4');
    grLfbConstantDepth := GetProcAddress (GlideHandle,  '_grLfbConstantDepth@4');
    grLfbWriteColorSwizzle := GetProcAddress (GlideHandle,  '_grLfbWriteColorSwizzle@8');
    grLfbWriteColorFormat := GetProcAddress (GlideHandle,  '_grLfbWriteColorFormat@4');
    grLfbWriteRegion := GetProcAddress (GlideHandle,  '_grLfbWriteRegion@32');
    grLfbReadRegion := GetProcAddress (GlideHandle,  '_grLfbReadRegion@28');

    //  Antialiasing Functions
    grAADrawLine := GetProcAddress (GlideHandle,  '_grAADrawLine@8');
    grAADrawPoint := GetProcAddress (GlideHandle,  '_grAADrawPoint@4');
    grAADrawPolygon := GetProcAddress (GlideHandle,  '_grAADrawPolygon@12');
    grAADrawPolygonVertexList := GetProcAddress (GlideHandle,  '_grAADrawPolygonVertexList@8');
    grAADrawTriangle := GetProcAddress (GlideHandle,  '_grAADrawTriangle@24');

    // glide management functions
    grGlideInit := GetProcAddress (GlideHandle,  '_grGlideInit@0');
    grGlideShutdown := GetProcAddress (GlideHandle,  '_grGlideShutdown@0');
    grGlideGetVersion := GetProcAddress (GlideHandle,  '_grGlideGetVersion@4');
    grGlideGetState := GetProcAddress (GlideHandle,  '_grGlideGetState@4');
    grGlideSetState := GetProcAddress (GlideHandle,  '_grGlideSetState@4');
    grGlideShamelessPlug := GetProcAddress (GlideHandle,  '_grGlideShamelessPlug@4');
    grHints := GetProcAddress (GlideHandle,  '_grHints@8');

{$IFDEF GLIDE3}
    grParameterData := GetProcAddress (GlideHandle,  '_grParameterData@16');
    grDrawArray := GetProcAddress (GlideHandle,  '_grDrawArray@12');
{$ENDIF}

// rendering functions
    guAADrawTriangleWithClip := GetProcAddress (GlideHandle,  '_guAADrawTriangleWithClip@12');
    guDrawTriangleWithClip := GetProcAddress (GlideHandle,  '_guDrawTriangleWithClip@12');
    guDrawPolygonVertexListWithClip := GetProcAddress (GlideHandle,  '_guDrawPolygonVertexListWithClip@8');

    // hi-level rendering utility functions
    guAlphaSource := GetProcAddress (GlideHandle,  '_guAlphaSource@4');
    guColorCombineFunction := GetProcAddress (GlideHandle,  '_guColorCombineFunction@4');
    guEncodeRLE16 := GetProcAddress (GlideHandle,  '_guEncodeRLE16@16');
    guTexCreateColorMipMap := GetProcAddress (GlideHandle,  '_guTexCreateColorMipMap@0');

    // fog stuff
    guFogTableIndexToW := GetProcAddress (GlideHandle,  '_guFogTableIndexToW@4');
    guFogGenerateExp := GetProcAddress (GlideHandle,  '_guFogGenerateExp@8');
    guFogGenerateExp2 := GetProcAddress (GlideHandle,  '_guFogGenerateExp2@8');
    guFogGenerateLinear := GetProcAddress (GlideHandle,  '_guFogGenerateLinear@12');

    // endian stuff
    guEndianSwapWords := GetProcAddress (GlideHandle,  '_guEndianSwapWords@4');
    guEndianSwapBytes := GetProcAddress (GlideHandle,  '_guEndianSwapBytes@4');

    // hi-level texture manipulation tools.
    gu3dfGetInfo := GetProcAddress (GlideHandle,  '_gu3dfGetInfo@8');
    gu3dfLoad := GetProcAddress (GlideHandle,  '_gu3dfLoad@8');

    // movie capture stuff
    guMovieStart := GetProcAddress (GlideHandle,  '_guMovieStart@0');
    guMovieStop := GetProcAddress (GlideHandle,  '_guMovieStop@0');
    guMovieSetName := GetProcAddress (GlideHandle,  '_guMovieSetName@4');

    Result:=True;
  end;
  GlideUnitIsInitalised := Result;
end;

function CloseGlideUnit : Boolean;
begin
  if GlideHandle <> 0 then
    begin
      FreeLibrary(GlideHandle);
      GlideHandle := 0;
    end;
  Result := True;
  GlideUnitIsInitalised := not Result;
end;

function GR_GENERATE_FIFOCHECK_HINT_MASK(swHWM, swLWM: variant) : variant;
begin
  Result := ((swHWM and $ffff) shl 16) or (swLWM and $ffff)
end;

function FXBIT(i: variant) : variant;
begin
  Result := 1 shl i;
end;

function FXUNUSED(a: variant) : boolean;
begin
  Result := (a = 0);
end;


procedure SetIntelPrecision;
var
	memvar : LongInt;
begin
	//taken directly from the Glide 2.4 programming Guide
	asm
		finit
		fwait
		fstcw word ptr memvar
		fwait
		mov eax,memvar
		and eax,0fffffcffh
		mov memvar,eax
		fldcw word ptr memvar
		fwait
	end;
end;

initialization
begin
  GlideHandle := 0;
  InitGlideUnit;

  {$IFNDEF NO_SETINTELPRECISION}
  SetIntelPrecision;
  {$ENDIF}
end

finalization
begin
  CloseGlideUnit;
end

end.
