// Pansen!
Unit MainForm;

Interface

Uses
   Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Glide2x, LFBProcs,
   MPlayer, MMSystem;

Const
   // Gre des Spielfelds
   MaxWidth = 7;
   MaxHeight = 13;
   // Gre der Steine
   BlockSize = 20;
   // Augenabstand
   EyeDistance = 768;
   // Sterne
   MaxStars = 2000;
   MaxStarsMainMenu = 249;
          // Spiel-Status
   gsMainMenu = 0;
   gsGame = 1;
   // Spiel-Modus
   gmLines = 0;
   gmPlanes = 1;
   // Schriften
   MaxFont = 0;
   // Steine
   MaxBlock = 6;
   LeftRightChangeXAxis = 0;
   LeftRightChangeZAxis = 1;
   // Textur(en)
   MaxTexture = 10;
   BlockTexOffset = 0; // Um 1 erniedrigt (eigentlich 1)
   MenuTexOffset = 8;
   MaxTextureBack = 0;
   // Maximale Anzahl der Eckpunkte
   MaxVertex = ((MaxWidth+1)*(MaxWidth+1)*(MaxHeight+1)*8)-1+8+((MaxWidth+1)+(MaxHeight+1)*6);
   // Vorschau auf nchsten Stein
   PreviewBlockPolygons = 20;
   PreviewBlockCenterX = 100;
   PreviewBlockCenterY = 100;
   PreviewBlockRadius = 100;
   // Men
   MenuItemRadius = 60;
   MenuItems = 6;
   MaxMenuVertecies = (MenuItems*4)-1;
   MaxMenuFaces = (MenuItems*2)-1;

Type
   TBlock = Record
      Block: Array[0..4, 0..4, 0..4] Of Byte;
      Width, Height, Depth: Byte;
      RotPoint: Record
          x, y, z: Byte;
      End;
      RotationNotAllowed: Boolean;
   End;

   TBlockPos = Record
      x, y, z: Integer;
      RotX, RotY, RotZ: Single;
   End;

   TPosIncFac = Record
      x, z: Integer;
      LeftRight: Integer;
   End;

   TIncFacPerMS = Record
      RotX, RotY: Single;
      Stars: Single;
      GameSpeed: Single;
   End;

   TBlockColor = Packed Record
      r, g, b: Single;
   End;

   TCharInfo = Record
      x, y, Width, Height: Integer;
   End;

   TFace = Record
      a, b, c: Longint;
   End;

   TFrmTetris = Class(TForm)
      AVIPlayer: TMediaPlayer;
      Procedure FormClose(Sender: TObject; Var Action: TCloseAction);
      Procedure FormCreate(Sender: TObject);
      procedure FormShow(Sender: TObject);
      procedure AVIPlayerNotify(Sender: TObject);
      procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
   Private
      bHardwareInitialized: Boolean;
      Procedure AppActivate(Sender: TObject);
      Procedure AppDeactivate(Sender: TObject);
      Function Enclosed(x, y, z: Word): Boolean;
      Procedure InsertBlock;
      Procedure RemoveBlock;
      Procedure RemoveCompleteLines;
      Procedure RemoveCompletePlanes;
      Procedure MoveBlockDown;
      Procedure InitNewBlock;
      Function TestBlock(a, b, c: Integer; RotX, RotY, RotZ: Single): Boolean;
      Procedure InitStars;
      Procedure DrawStars;
      Procedure MoveStars;
      Procedure InitStarsMainMenu;
      Procedure InitFont(FontNr: Integer; FontFile: String);
      Procedure OutTextXY(x, y, FontNr, FontSpacing: Integer; Text: String);
      Procedure CreateShadow;
      Procedure InitPreviewBlockBackground;
      Procedure DrawPreviewBlockBackground;
      Procedure DrawPreviewBlock;
      Procedure SetEntry(Entry, r, g, b: Byte);
      Procedure SetPalette;
      Procedure InitMenuItems;
      Procedure DrawMainMenu;
      Procedure DrawGame;
      Procedure DrawOnIdle(Sender: TObject; Var Done: Boolean);
   Public
      aField: Array[0..MaxWidth, 0..MaxWidth, 0..MaxHeight] Of Byte;
      aVertecies: Array[0..MaxVertex] Of TgrVertex;
      aTexIndecies: Array[0..MaxVertex Div 8] Of Integer;
      aPreviewVertecies: Array[0..PreviewBlockPolygons] Of TgrVertex;
      aMenuVertecies: Array[0..MaxMenuVertecies] Of TgrVertex;
      aMenuFaces: Array[0..MaxMenuFaces] Of TFace;
      aStars: Array[0..MaxStars] Of Record
         x, y, z: Single;
         xscr, yscr: Integer;
         Size: Integer;
         Color: Word;
      End;
      aStarsMainMenu: Array[0..MaxStarsMainMenu] Of Record
         x, y: Integer;
         Plane: Integer;
         xInc: Integer;
         Size: Integer;
      End;
      aFonts: Array[0..MaxFont, 0..255] Of LFB_Sprite;
      aFontCoordinates: Array[0..MaxFont, 0..255] Of TCharInfo;
      aFontFiles: Array[0..MaxFont] Of String;
      aBlockColors: Array[0..MaxBlock] Of TBlockColor;
      aBlocks: Array[0..MaxBlock] Of TBlock;
      CurrentBlock: TBlock;
      BlockPreview: TBlock;
      iVerteciesNum: Integer;
      iTexIndeciesNum: Integer;
      iGridVerteciesNum: Integer;
      iShadowVerteciesNum: Integer;
      sgRotX, sgRotY, sgRotZ: Single;
      sgRotXp, sgRotYp, sgRotZp: Single;
      sgRotXm, sgRotYm, sgRotZm: Single;
      iZoom: Integer;
      iGameState: Integer;
      iActBlock: Integer;
      iNextBlock: Integer;
      iStarsSpeed: Integer;
      iGameSpeed: Integer;
      sgTickCounter: Single;
      BlockPos: TBlockPos;
      PosIncFac: TPosIncFac;
      IncFacPerMS: TIncFacPerMS;
      iFrameTime: Integer;
      aTextures: Array[0..MaxTexture] Of TgrTexInfo;
      aTextureAddresses: Array[0..MaxTexture] Of Longint;
      iTextureNum: Integer;
      aTexturesBack: Array[0..MaxTextureBack] Of TgrTexInfo;
      aTextureAddressesBack: Array[0..MaxTextureBack] Of Longint;
      iTextureNumBack: Integer;
      sAppPath: String;
      iGameMode: Integer;
      iRotMenu: Integer;

      Procedure InitVar;
      Procedure InitHardware;
      Procedure ShutDownHardware;
   End;

Var
   FrmTetris: TFrmTetris;
//   DDInterface: IDirectDraw2;
   fps: Single;
   Phase: Word;
   Palette: Array[0..32767] Of Word;
   Plasma: LFB_Sprite;
   MenuBar, MenuLogo: LFB_Sprite;

Implementation

{$R *.DFM}

Function Stri(i: Longint): String;
Begin
     Str(i, Result);
End;

Procedure TFrmTetris.InitVar;
Var
   Dummy: LFB_Sprite;
//   x, y, z: Integer;
   i: Integer;
Begin
     SetPalette;
     bHardwareInitialized:=False;
     iGameSpeed:=500;
     sgTickCounter:=0;
     InitStars;
     InitStarsMainMenu;
     FillChar(aField, SizeOf(aField), #0);
     FillChar(aVertecies, SizeOf(aVertecies), #0);
     FillChar(aBlocks, SizeOf(aBlocks), #0);
{     For x:=0 To MaxWidth Do For y:=0 To MaxHeight Do For z:=0 To MaxWidth Do
         aField[x, z, y]:=Random(MaxBlock);}
     sgRotX:=Pi/4;
     sgRotY:=Pi/8;
     sgRotZ:=0;
     sgRotXp:=Pi/4;
     sgRotYp:=0;
     sgRotZp:=Pi;
     sgRotXm:=Pi/12;
     sgRotYm:=0;
     sgRotZm:=0;
     iZoom:=400;
     iRotMenu:=0;

     FillChar(CurrentBlock, SizeOf(CurrentBlock), #0);
     With CurrentBlock Do Begin
          Width:=5;
          Height:=5;
          Depth:=5;
     End;

     With IncFacPerMS Do Begin
          RotX:=Pi/1000;
          RotY:=Pi/1000;
          Stars:=1;
          GameSpeed:=1;
     End;

     iGameState:=gsGame; // Nur vorlufig...
     iGameMode:=gmLines;

     aBlockColors[0].r:=0;
     aBlockColors[0].g:=0;
     aBlockColors[0].b:=255;

     aBlockColors[1].r:=0;
     aBlockColors[1].g:=255;
     aBlockColors[1].b:=0;

     aBlockColors[2].r:=255;
     aBlockColors[2].g:=0;
     aBlockColors[2].b:=0;

     aBlockColors[3].r:=255;
     aBlockColors[3].g:=255;
     aBlockColors[3].b:=0;

     aBlockColors[4].r:=255;
     aBlockColors[4].g:=0;
     aBlockColors[4].b:=255;

     aBlockColors[5].r:=0;
     aBlockColors[5].g:=255;
     aBlockColors[5].b:=255;

     aBlockColors[6].r:=125;
     aBlockColors[6].g:=125;
     aBlockColors[6].b:=0;

{     aBlockColors[7].r:=0;
     aBlockColors[7].g:=125;
     aBlockColors[7].b:=125;}

{     aBlockColors[8].r:=125;
     aBlockColors[8].g:=0;
     aBlockColors[8].b:=125;

     aBlockColors[9].r:=0;
     aBlockColors[9].g:=125;
     aBlockColors[9].b:=0;

     aBlockColors[10].r:=125;
     aBlockColors[10].g:=0;
     aBlockColors[10].b:=0;

     aBlockColors[11].r:=0;
     aBlockColors[11].g:=0;
     aBlockColors[11].b:=125;

     aBlockColors[12].r:=255;
     aBlockColors[12].g:=0;
     aBlockColors[12].b:=125;

     aBlockColors[13].r:=125;
     aBlockColors[13].g:=255;
     aBlockColors[13].b:=0;

     aBlockColors[14].r:=125;
     aBlockColors[14].g:=0;
     aBlockColors[14].b:=255;

     aBlockColors[15].r:=255;
     aBlockColors[15].g:=125;
     aBlockColors[15].b:=255;

     aBlockColors[16].r:=125;
     aBlockColors[16].g:=255;
     aBlockColors[16].b:=125;}


     aBlocks[0].Block[0, 0, 0]:=1;
     aBlocks[0].Block[0, 1, 0]:=1;
     aBlocks[0].Block[0, 2, 0]:=1;
     aBlocks[0].Block[0, 3, 0]:=1;
     aBlocks[0].Width:=1;
     aBlocks[0].Height:=4;
     aBlocks[0].Depth:=1;
     aBlocks[0].RotPoint.x:=0;
     aBlocks[0].RotPoint.y:=1;
     aBlocks[0].RotPoint.z:=0;


     aBlocks[1].Block[0, 0, 0]:=2;
     aBlocks[1].Block[0, 1, 0]:=2;
     aBlocks[1].Block[0, 2, 0]:=2;
     aBlocks[1].Block[0, 3, 0]:=2;
     aBlocks[1].Block[1, 1, 0]:=2;
     aBlocks[1].Width:=2;
     aBlocks[1].Height:=4;
     aBlocks[1].Depth:=1;
     aBlocks[1].RotPoint.x:=0;
     aBlocks[1].RotPoint.y:=1;
     aBlocks[1].RotPoint.z:=0;

     aBlocks[2].Block[1, 0, 0]:=3;
     aBlocks[2].Block[0, 1, 0]:=3;
     aBlocks[2].Block[1, 1, 0]:=3;
     aBlocks[2].Block[2, 1, 0]:=3;
     aBlocks[2].Width:=3;
     aBlocks[2].Height:=2;
     aBlocks[2].Depth:=1;
     aBlocks[2].RotPoint.x:=1;
     aBlocks[2].RotPoint.y:=1;
     aBlocks[2].RotPoint.z:=0;

     aBlocks[3].Block[0, 0, 0]:=4;
     aBlocks[3].Block[0, 1, 0]:=4;
     aBlocks[3].Block[1, 1, 0]:=4;
     aBlocks[3].Width:=2;
     aBlocks[3].Height:=2;
     aBlocks[3].Depth:=1;
     aBlocks[3].RotPoint.x:=0;
     aBlocks[3].RotPoint.y:=1;
     aBlocks[3].RotPoint.z:=0;

     aBlocks[4].Block[0, 0, 0]:=5;
     aBlocks[4].Block[1, 0, 0]:=5;
     aBlocks[4].Block[1, 1, 0]:=5;
     aBlocks[4].Block[2, 1, 0]:=5;
     aBlocks[4].Width:=3;
     aBlocks[4].Height:=2;
     aBlocks[4].Depth:=1;
     aBlocks[4].RotPoint.x:=1;
     aBlocks[4].RotPoint.y:=1;
     aBlocks[4].RotPoint.z:=0;

     aBlocks[5].Block[0, 0, 0]:=6;
     aBlocks[5].Block[1, 0, 0]:=6;
     aBlocks[5].Block[0, 1, 0]:=6;
     aBlocks[5].Block[1, 1, 0]:=6;
     aBlocks[5].Block[0, 0, 1]:=6;
     aBlocks[5].Block[1, 0, 1]:=6;
     aBlocks[5].Block[0, 1, 1]:=6;
     aBlocks[5].Block[1, 1, 1]:=6;
     aBlocks[5].Width:=2;
     aBlocks[5].Height:=2;
     aBlocks[5].Depth:=2;
     aBlocks[5].RotPoint.x:=0;
     aBlocks[5].RotPoint.y:=0;
     aBlocks[5].RotPoint.z:=0;
     aBlocks[5].RotationNotAllowed:=True;

     aBlocks[6].Block[0, 0, 0]:=7;
     aBlocks[6].Block[0, 1, 0]:=7;
     aBlocks[6].Block[0, 2, 0]:=7;
     aBlocks[6].Block[1, 2, 0]:=7;
     aBlocks[6].Width:=2;
     aBlocks[6].Height:=3;
     aBlocks[6].Depth:=1;
     aBlocks[6].RotPoint.x:=0;
     aBlocks[6].RotPoint.y:=2;
     aBlocks[6].RotPoint.z:=0;

{     aBlocks[7].Block[1, 0, 0]:=8;
     aBlocks[7].Block[1, 1, 0]:=8;
     aBlocks[7].Block[0, 2, 0]:=8;
     aBlocks[7].Block[1, 2, 0]:=8;
     aBlocks[7].Block[2, 2, 0]:=8;
     aBlocks[7].Width:=3;
     aBlocks[7].Height:=3;
     aBlocks[7].Depth:=1;
     aBlocks[7].RotPoint.x:=1;
     aBlocks[7].RotPoint.y:=2;
     aBlocks[7].RotPoint.z:=0;}

{     aBlocks[2].Block[0, 0, 0]:=3;
     aBlocks[2].Block[0, 1, 0]:=3;
     aBlocks[2].Block[0, 2, 0]:=3;
     aBlocks[2].Block[0, 3, 0]:=3;
     aBlocks[2].Block[1, 1, 0]:=3;
     aBlocks[2].Block[0, 1, 1]:=3;
     aBlocks[2].Width:=2;
     aBlocks[2].Height:=4;
     aBlocks[2].Depth:=2;
     aBlocks[2].RotPoint.x:=0;
     aBlocks[2].RotPoint.y:=1;
     aBlocks[2].RotPoint.z:=0;

     aBlocks[3].Block[1, 0, 0]:=4;
     aBlocks[3].Block[1, 1, 0]:=4;
     aBlocks[3].Block[1, 2, 0]:=4;
     aBlocks[3].Block[1, 3, 0]:=4;
     aBlocks[3].Block[0, 1, 0]:=4;
     aBlocks[3].Block[2, 1, 0]:=4;
     aBlocks[3].Block[1, 1, 1]:=4;
     aBlocks[3].Width:=3;
     aBlocks[3].Height:=4;
     aBlocks[3].Depth:=2;
     aBlocks[3].RotPoint.x:=1;
     aBlocks[3].RotPoint.y:=1;
     aBlocks[3].RotPoint.z:=0;

     aBlocks[4].Block[1, 0, 1]:=5;
     aBlocks[4].Block[1, 1, 1]:=5;
     aBlocks[4].Block[1, 2, 1]:=5;
     aBlocks[4].Block[1, 3, 1]:=5;
     aBlocks[4].Block[0, 1, 1]:=5;
     aBlocks[4].Block[1, 1, 0]:=5;
     aBlocks[4].Block[2, 1, 1]:=5;
     aBlocks[4].Block[1, 1, 2]:=5;
     aBlocks[4].Width:=3;
     aBlocks[4].Height:=4;
     aBlocks[4].Depth:=3;
     aBlocks[4].RotPoint.x:=1;
     aBlocks[4].RotPoint.y:=1;
     aBlocks[4].RotPoint.z:=1;

     aBlocks[5].Block[1, 0, 0]:=6;
     aBlocks[5].Block[1, 1, 0]:=6;
     aBlocks[5].Block[1, 2, 0]:=6;
     aBlocks[5].Block[1, 3, 0]:=6;
     aBlocks[5].Block[0, 1, 0]:=6;
     aBlocks[5].Block[2, 1, 0]:=6;
     aBlocks[5].Width:=3;
     aBlocks[5].Height:=4;
     aBlocks[5].Depth:=1;
     aBlocks[5].RotPoint.x:=1;
     aBlocks[5].RotPoint.y:=1;
     aBlocks[5].RotPoint.z:=0;

     aBlocks[8].Block[1, 0, 0]:=9;
     aBlocks[8].Block[0, 1, 0]:=9;
     aBlocks[8].Block[1, 1, 0]:=9;
     aBlocks[8].Block[2, 1, 0]:=9;
     aBlocks[8].Block[1, 1, 1]:=9;
     aBlocks[8].Width:=3;
     aBlocks[8].Height:=2;
     aBlocks[8].Depth:=2;
     aBlocks[8].RotPoint.x:=1;
     aBlocks[8].RotPoint.y:=1;
     aBlocks[8].RotPoint.z:=0;

     aBlocks[9].Block[1, 0, 1]:=10;
     aBlocks[9].Block[0, 1, 1]:=10;
     aBlocks[9].Block[1, 1, 1]:=10;
     aBlocks[9].Block[2, 1, 1]:=10;
     aBlocks[9].Block[1, 1, 0]:=10;
     aBlocks[9].Block[1, 1, 2]:=10;
     aBlocks[9].Width:=3;
     aBlocks[9].Height:=2;
     aBlocks[9].Depth:=3;
     aBlocks[9].RotPoint.x:=1;
     aBlocks[9].RotPoint.y:=1;
     aBlocks[9].RotPoint.z:=1;

     aBlocks[13].Block[0, 0, 0]:=14;
     aBlocks[13].Block[0, 1, 0]:=14;
     aBlocks[13].Block[0, 2, 0]:=14;
     aBlocks[13].Block[1, 2, 0]:=14;
     aBlocks[13].Block[0, 2, 1]:=14;
     aBlocks[13].Width:=2;
     aBlocks[13].Height:=3;
     aBlocks[13].Depth:=2;
     aBlocks[13].RotPoint.x:=0;
     aBlocks[13].RotPoint.y:=2;
     aBlocks[13].RotPoint.z:=0;

     aBlocks[15].Block[1, 0, 0]:=16;
     aBlocks[15].Block[1, 1, 0]:=16;
     aBlocks[15].Block[0, 2, 0]:=16;
     aBlocks[15].Block[1, 2, 0]:=16;
     aBlocks[15].Block[2, 2, 0]:=16;
     aBlocks[15].Block[1, 2, 1]:=16;
     aBlocks[15].Width:=3;
     aBlocks[15].Height:=3;
     aBlocks[15].Depth:=2;
     aBlocks[15].RotPoint.x:=0;
     aBlocks[15].RotPoint.y:=2;
     aBlocks[15].RotPoint.z:=0;

     aBlocks[16].Block[1, 2, 0]:=17;
     aBlocks[16].Block[1, 0, 1]:=17;
     aBlocks[16].Block[1, 1, 1]:=17;
     aBlocks[16].Block[0, 2, 1]:=17;
     aBlocks[16].Block[1, 2, 1]:=17;
     aBlocks[16].Block[2, 2, 1]:=17;
     aBlocks[16].Block[1, 2, 2]:=17;
     aBlocks[16].Width:=3;
     aBlocks[16].Height:=3;
     aBlocks[16].Depth:=3;
     aBlocks[16].RotPoint.x:=0;
     aBlocks[16].RotPoint.y:=2;
     aBlocks[16].RotPoint.z:=1;}

     iNextBlock:=Random(MaxBlock+1);
     InitNewBlock;

     FillChar(aFonts, SizeOf(aFonts), #0);
     FillChar(aFontCoordinates, SizeOf(aFontCoordinates), #0);

     aFontCoordinates[0,65].x:=15;
     aFontCoordinates[0,65].y:=12;
     aFontCoordinates[0,65].Width:=20;
     aFontCoordinates[0,65].Height:=29;

     aFontCoordinates[0,66].x:=39;
     aFontCoordinates[0,66].y:=12;
     aFontCoordinates[0,66].Width:=18;
     aFontCoordinates[0,66].Height:=29;

     aFontCoordinates[0,67].x:=63;
     aFontCoordinates[0,67].y:=11;
     aFontCoordinates[0,67].Width:=19;
     aFontCoordinates[0,67].Height:=31;

     aFontCoordinates[0,68].x:=87;
     aFontCoordinates[0,68].y:=12;
     aFontCoordinates[0,68].Width:=19;
     aFontCoordinates[0,68].Height:=29;

     aFontCoordinates[0,48].x:=602;
     aFontCoordinates[0,48].y:=11;
     aFontCoordinates[0,48].Width:=18;
     aFontCoordinates[0,48].Height:=31;

     aFontCoordinates[0,49].x:=624;
     aFontCoordinates[0,49].y:=12;
     aFontCoordinates[0,49].Width:=13;
     aFontCoordinates[0,49].Height:=29;

     aFontCoordinates[0,50].x:=643;
     aFontCoordinates[0,50].y:=11;
     aFontCoordinates[0,50].Width:=17;
     aFontCoordinates[0,50].Height:=30;

     aFontCoordinates[0,51].x:=665;
     aFontCoordinates[0,51].y:=11;
     aFontCoordinates[0,51].Width:=18;
     aFontCoordinates[0,51].Height:=31;

     aFontCoordinates[0,52].x:=688;
     aFontCoordinates[0,52].y:=12;
     aFontCoordinates[0,52].Width:=18;
     aFontCoordinates[0,52].Height:=30;

     aFontCoordinates[0,53].x:=711;
     aFontCoordinates[0,53].y:=12;
     aFontCoordinates[0,53].Width:=18;
     aFontCoordinates[0,53].Height:=30;

     aFontCoordinates[0,54].x:=735;
     aFontCoordinates[0,54].y:=11;
     aFontCoordinates[0,54].Width:=18;
     aFontCoordinates[0,54].Height:=30;

     aFontCoordinates[0,55].x:=757;
     aFontCoordinates[0,55].y:=12;
     aFontCoordinates[0,55].Width:=15;
     aFontCoordinates[0,55].Height:=29;

     aFontCoordinates[0,56].x:=776;
     aFontCoordinates[0,56].y:=11;
     aFontCoordinates[0,56].Width:=18;
     aFontCoordinates[0,56].Height:=31;

     aFontCoordinates[0,57].x:=800;
     aFontCoordinates[0,57].y:=11;
     aFontCoordinates[0,57].Width:=18;
     aFontCoordinates[0,57].Height:=31;

{     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;

     aFontCoordinates[].x:=;
     aFontCoordinates[].y:=;
     aFontCoordinates[].Width:=;
     aFontCoordinates[].Height:=;
}

     LFB_LoadPCXFile(sAppPath+'GFX\Carrier.pcx', Dummy);
     With aTextures[0] Do Begin
          Format:=GR_TEXFMT_RGB_565;
          aspectRatio:=GR_ASPECT_1x1;
          smallLod:=GR_LOD_256;
          largeLod:=GR_LOD_256;
          data:=Dummy.Adr;
     End;
     iTextureNum:=1;
     For i:=1 To MaxBlock+1 Do Begin
         LFB_LoadPCXFile(sAppPath+'GFX\btex'+Stri(i)+'.pcx', Dummy);
         With aTextures[i] Do Begin
              Format:=GR_TEXFMT_RGB_565;
              aspectRatio:=GR_ASPECT_1x1;
              smallLod:=GR_LOD_32;
              largeLod:=GR_LOD_32;
              data:=Dummy.Adr;
         End;
         Inc(iTextureNum);
     End;
     LFB_LoadPCXFile(sAppPath+'GFX\mi1.pcx', Dummy);
     With aTextures[iTextureNum] Do Begin
          Format:=GR_TEXFMT_RGB_565;
          aspectRatio:=GR_ASPECT_4x1;
          smallLod:=GR_LOD_256;
          largeLod:=GR_LOD_256;
          data:=Dummy.Adr;
     End;
     Inc(iTextureNum);

     LFB_LoadPCXFile(sAppPath+'GFX\mb1.pcx', Dummy);
     With aTexturesBack[0] Do Begin
          Format:=GR_TEXFMT_RGB_565;
          aspectRatio:=GR_ASPECT_4x1;
          smallLod:=GR_LOD_256;
          largeLod:=GR_LOD_256;
          data:=Dummy.Adr;
     End;
     iTextureNumBack:=1;

     InitFont(0, sAppPath+'GFX\Schrift.pcx');
     InitPreviewBlockBackground;
     InitMenuItems;
     With Plasma Do Begin
          Width:=640;
          Height:=480;
          GetMem(Adr, Width*Height*2);
     End;
     LFB_LoadPCXFile(sAppPath+'GFX\Menu.pcx', MenuBar);
     LFB_LoadPCXFile(sAppPath+'GFX\Logo.pcx', MenuLogo);
End;

Procedure TFrmTetris.InitHardware;
Var
   StartAddress: Longint;
   i: Integer;
Begin
     If Not bHardwareInitialized Then Begin
        grGlideInit;
        grSstSelect(0);
        If Not grSstWinOpen(FrmTetris.Handle, GR_RESOLUTION_1024X768, GR_REFRESH_60HZ, GR_COLORFORMAT_RGBA, GR_ORIGIN_UPPER_LEFT, 2, 1) Then Begin
           grGlideShutDown;
           MessageDlg('No 3dfx Voodoo found.', mtError, [mbOK], 0);
           Halt;
        End;
        grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
        grDepthBufferFunction(GR_CMP_LESS);
        grDepthMask(FXTRUE);
        grCullMode(GR_CULL_POSITIVE);
        grBufferClear(0,0,$FFFF);
        grBufferSwap(0);
        StartAddress:=grTexMinAddress(GR_TMU0);
        For i:=0 To iTextureNum-1 Do Begin
            aTextureAddresses[i]:=StartAddress;
            grTexDownLoadMipMap(GR_TMU0, StartAddress, GR_MIPMAPLEVELMASK_BOTH, aTextures[i]);
            Inc(StartAddress, grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH, aTextures[i]));
        End;
        StartAddress:=grTexMinAddress(GR_TMU1);
        For i:=0 To iTextureNumBack-1 Do Begin
            aTextureAddressesBack[i]:=StartAddress;
            grTexDownLoadMipMap(GR_TMU1, StartAddress, GR_MIPMAPLEVELMASK_BOTH, aTexturesBack[i]);
            Inc(StartAddress, grTexTextureMemRequired(GR_MIPMAPLEVELMASK_BOTH, aTexturesBack[i]));
        End;
        guColorCombineFunction(GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB);
        LFB_SetClippingRect(0, 0, 1023, 767);
        grTexFilterMode(GR_TMU0, GR_TEXTUREFILTER_BILINEAR, GR_TEXTUREFILTER_BILINEAR);
        grTexMipMapMode(GR_TMU0, GR_MIPMAP_DISABLE, FXFALSE);
        grTexFilterMode(GR_TMU1, GR_TEXTUREFILTER_BILINEAR, GR_TEXTUREFILTER_BILINEAR);
        grTexMipMapMode(GR_TMU1, GR_MIPMAP_DISABLE, FXFALSE);
        grTexCombine(GR_TMU0, GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE,
                     GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE, FXFALSE, FXFALSE);
        bHardwareInitialized:=True;
     End;
End;

Procedure TFrmTetris.ShutDownHardware;
Begin
     If bHardwareInitialized Then Begin
        grSstWinClose;
        grGlideShutDown;
        LFB_SetClippingRect(0, 0, 0, 0);
        bHardwareInitialized:=False;
     End;
End;

Procedure TFrmTetris.FormClose(Sender: TObject; Var Action: TCloseAction);
Begin
     ShutDownHardware;
End;

Procedure TFrmTetris.AppActivate;
Begin
     InitHardware;
     Application.OnIdle:=DrawOnIdle;
End;

Procedure TFrmTetris.AppDeactivate;
Begin
     Application.OnIdle:=Nil;
     ShutDownHardware;
End;

Procedure TFrmTetris.FormCreate(Sender: TObject);
Begin
     sAppPath:=ExtractFilePath(ParamStr(0));
     InitVar;
     Application.OnActivate:=AppActivate;
     Application.OnDeactivate:=AppDeactivate;
End;

procedure TFrmTetris.FormShow(Sender: TObject);
{Var
   DD1: IDirectDraw;
   SD: TDDSurfaceDesc;
   SC: TDDSCAPS;}
begin
{     If Not Succeeded(DirectDrawCreate(Nil, DD1, Nil)) Then Begin
        MessageDlg('Failed to create DirectDraw-Interface.', mtError, [mbOK], 0);
        Halt;
     End;
     DD1._AddRef;
     If Not Succeeded(DD1.QueryInterface(IID_IDirectDraw2, DDInterface))Then Begin
        MessageDlg('Failed to create DirectDraw-Interface.', mtError, [mbOK], 0);
        Halt;
     End;
     DD1._Release;
     DDInterface._AddRef;
     If Not Succeeded(DDInterface.SetCooperativeLevel(Handle, DDSCL_EXCLUSIVE+DDSCL_FULLSCREEN+DDSCL_ALLOWMODEX+DDSCL_ALLOWREBOOT)) Then Begin
        MessageDlg('SetCooperativeLevel failed.', mtError, [mbOK], 0);
        Halt;
     End;
     If Not Succeeded(DDInterface.SetDisplayMode(320, 200, 16, 0, 0)) Then Begin
        MessageDlg('SetDisplayMode (320x200x16) failed.', mtError, [mbOK], 0);
        Halt;
     End;
     Screen.Cursor:=crNone;
     AVIPlayer.Open;
     AVIPlayer.Play;}
end;

procedure TFrmTetris.AVIPlayerNotify(Sender: TObject);
begin
{     Screen.Cursor:=crDefault;
     If Assigned(DDInterface) Then Begin
        DDInterface.FlipToGDISurface;
        DDInterface.RestoreDisplayMode;
        DDInterface._Release;
     End;
     Application.OnActivate:=AppActivate;
     Application.OnDeactivate:=AppDeactivate;
     AppActivate(Nil);}
end;

Procedure TFrmTetris.DrawOnIdle;
Begin
     If Assigned(Application.OnIdle) Then Begin
{              grGlideShutDown;
              MessageDlg('No 3dfx Voodoo found.', mtError, [mbOK], 0);
              Halt;}
        Done:=False;
        Case iGameState Of
//             gsMainMenu: DrawMainMenu;
             gsGame: DrawGame;
        End;
     End;
End;

Procedure TFrmTetris.InitStars;
Var
   i: Integer;
   RGB: Byte;
Begin
     Randomize;
     For i:=0 To MaxStars Do With aStars[i] Do Begin
         x:=Random(10000)-5000;
         y:=Random(10000)-5000;
         z:=Random(10000)+100;
         RGB:=Random(255);
         Color:=LFB_Encode565(RGB Shr 3, RGB Shr 2, RGB Shr 3);
         Size:=Round(Random(75)/25)+1;
     End;
End;

Procedure TFrmTetris.DrawStars;
Var
   i: Integer;
Begin
     If LFB_Lock <> Nil Then Begin
        For i:=0 To MaxStars Do With aStars[i] Do Begin
            If (xscr >= cr_x1+1) And (yscr >= cr_y1+1) And (xscr <= cr_x2-1) And (yscr <= cr_y2-1) Then
               Case Size Of
                    1:LFB_PutPixel(xscr, yscr, Color);
                    2:Begin
                           LFB_PutPixel(xscr, yscr, Color);
                           LFB_PutPixel(xscr+1, yscr, Color);
                    End;
                    3:Begin
                           LFB_PutPixel(xscr, yscr, Color);
                           LFB_PutPixel(xscr+1, yscr, Color);
                           LFB_PutPixel(xscr, yscr+1, Color);
                    End;
                    4:Begin
                           LFB_PutPixel(xscr, yscr, Color);
                           LFB_PutPixel(xscr+1, yscr, Color);
                           LFB_PutPixel(xscr, yscr+1, Color);
                           LFB_PutPixel(xscr+1, yscr+1, Color);
                    End;
               End;
        End;
     End;
     LFB_Unlock;
End;

Procedure TFrmTetris.MoveStars;
Var
   i: Integer;
   RGB: Byte;
Begin
     For i:=0 To MaxStars Do With aStars[i] Do Begin
         If z > 0 Then Begin
            xscr:=Round((EyeDistance*x/z) + 512);
            yscr:=Round((EyeDistance*y/z) + 384);
         End;
         If (z < 10) Or (xscr < cr_x1+1) Or (yscr < cr_y1+1) Or (xscr > cr_x2-1) Or (yscr > cr_y2-1) Then Begin
            x:=Random(10000)-5000;
            y:=Random(10000)-5000;
            z:=10000;
            RGB:=Random(255);
            Color:=LFB_Encode565(RGB Shr 3, RGB Shr 2, RGB Shr 3);
            Size:=Round(Random(75)/25)+1;
         End;
         z:=z-iStarsSpeed;
     End;
End;

Function TFrmTetris.Enclosed;
Begin
     If (x = 0) Or (x = MaxWidth) Or (y = 0) Or (y = MaxHeight) Or (z = 0) Or (z = MaxWidth) Then
        Result:=False
     Else Begin
        If (aField[x-1, z, y] > 0) And
           (aField[x+1, z, y] > 0) And
           (aField[x, z-1, y] > 0) And
           (aField[x, z+1, y] > 0) And
           (aField[x, z, y+1] > 0) And
           (aField[x, z, y-1] > 0) Then Result:=True Else Result:=False;
     End;
End;

Procedure TFrmTetris.InsertBlock;
Var
   x, y, z: Integer;
Begin
     For x:=0 To CurrentBlock.Width-1 Do Begin
         For y:=0 To CurrentBlock.Height-1 Do Begin
             For z:=0 To CurrentBlock.Depth-1 Do Begin
                 If CurrentBlock.Block[x, y, z] > 0 Then
                    If y+BlockPos.y-2 <= MaxHeight Then
                       aField[x+BlockPos.x-2, z+BlockPos.z-2, y+BlockPos.y-2]:=CurrentBlock.Block[x, y, z];
             End;
         End;
     End;
End;

Procedure TFrmTetris.RemoveBlock;
Var
   x, y, z: Integer;
Begin
     For x:=0 To CurrentBlock.Width-1 Do Begin
         For y:=0 To CurrentBlock.Height-1 Do Begin
             For z:=0 To CurrentBlock.Depth-1 Do Begin
                 If CurrentBlock.Block[x, y, z] > 0 Then
                    If y+BlockPos.y-2 <= MaxHeight Then
                       aField[x+BlockPos.x-2, z+BlockPos.z-2, y+BlockPos.y-2]:=0;
             End;
         End;
     End;
End;

Function TFrmTetris.TestBlock;
Var
   a0, b0, c0: Integer;
   a1, b1, c1: Single;
   a2, b2, c2: Single;
   x, y, z: Integer;
   Tmp, Tmp2: TBlock;
   TmpEbene: Array[0..5, 0..5] Of Byte;
   sx, cx, sy, cy, sz, cz: Single;
Begin
     Result:=True;

     Tmp.Width:=5;
     Tmp.Height:=5;
     Tmp.Depth:=5;
     FillChar(Tmp.Block, Sizeof(Tmp.Block), #0);
     Tmp2:=Tmp;

     sx:=sin(RotX);
     sy:=sin(RotY);
     sz:=sin(RotZ);
     cx:=cos(RotX);
     cy:=cos(RotY);
     cz:=cos(RotZ);

     a0:=2-aBlocks[iActBlock].RotPoint.x;
     b0:=2-aBlocks[iActBlock].RotPoint.y;
     c0:=2-aBlocks[iActBlock].RotPoint.z;

     For x:=0 To aBlocks[iActBlock].Width-1 Do Begin
         For y:=0 To aBlocks[iActBlock].Height-1 Do Begin
             For z:=0 To aBlocks[iActBlock].Depth-1 Do Begin
                 Tmp2.Block[a0+x, b0+y, c0+z]:=aBlocks[iActBlock].Block[x, y, z];
             End;
         End;
     End;

     If aBlocks[iActBlock].RotationNotAllowed Then
        Tmp.Block:=Tmp2.Block
     Else Begin
        For x:=-2 To 2 Do Begin
            For y:=-2 To 2 Do Begin
                For z:=-2 To 2 Do Begin
                    If Tmp2.Block[x+2, y+2, z+2] > 0 Then Begin

                       // X-Achse
                       a1:=x;
                       b1:=(y*cx)-(z*sx);
                       c1:=(y*sx)+(z*cx);

                       // Drehung um Y-Achse
                       a2:=(a1*cy)+(c1*sy);
                       b2:=b1;
                       c2:=-(a1*sy)+(c1*cy);

                       // Drehung um Z-Achse
                       a0:=Round((a2*cz)-(b2*sz));
                       b0:=Round((a2*sz)+(b2*cz));
                       c0:=Round(c2);

                       Tmp.Block[a0+2, b0+2, c0+2]:=Tmp2.Block[x+2, y+2, z+2];

                    End;
                End;
            End;
        End;
     End;

     For x:=0 To Tmp.Width-1 Do Begin
         For y:=0 To Tmp.Height-1 Do Begin
             For z:=0 To Tmp.Depth-1 Do Begin
                 If Tmp.Block[x, y, z] > 0 Then Begin
                    a0:=x+a-2;
                    b0:=y+b-2;
                    c0:=z+c-2;
                    If (a0 < 0) Or (a0 > MaxWidth) Or (b0 < 0) Or
                       (c0 < 0) Or (c0 > MaxWidth) Then Begin
                          Result:=False;
                          Exit;
                    End;
                    If b0 <= MaxHeight Then
                       If (aField[a0, c0, b0] > 0) Then Begin
                          Result:=False;
                          Exit;
                       End;
                 End;
             End;
         End;
     End;

     CurrentBlock:=Tmp;
     BlockPos.x:=a;
     BlockPos.y:=b;
     BlockPos.z:=c;
     BlockPos.RotX:=RotX;
     BlockPos.RotY:=RotY;
     BlockPos.RotZ:=RotZ;
End;

Procedure TFrmTetris.RemoveCompleteLines;
Var
   x, y, z: Integer;
   i: Integer;
   Complete: Array[0..MaxHeight] Of Boolean;
   Tmp: Array[0..MaxWidth, 0..MaxHeight] Of Byte;
Begin
     // Alle vertikalen Ebenen durchgehen
     For z:=0 To MaxWidth Do Begin
         // Jede Zeile in der Ebene auf Vollstndigkeit berprfen
         For y:=0 To MaxHeight Do Begin
             Complete[y]:=True;
             For x:=0 To MaxWidth Do Begin
                 If aField[x, z, y] = 0 Then Begin
                    Complete[y]:=False;
                    Break;
                 End;
             End;
         End;
         // Alle vollstndigen Zeilen entfernen
         i:=0;
         FillChar(Tmp, SizeOf(Tmp), #0);
         For y:=0 To MaxHeight Do Begin
             If Not Complete[y] Then Begin
                For x:=0 To MaxWidth Do Tmp[x, i]:=aField[x, z, y];
                Inc(i);
             End;
         End;
         For x:=0 To MaxWidth Do For y:=0 To MaxHeight Do aField[x, z, y]:=Tmp[x, y];
     End;
End;

Procedure TFrmTetris.RemoveCompletePlanes;
Var
   x, y, z: Integer;
   i: Integer;
   Ok: Boolean;
   Complete: Array[0..MaxHeight] Of Boolean;
   Tmp: Array[0..MaxWidth, 0..MaxWidth, 0..MaxHeight] Of Byte;
Begin
     // Alle Horizontalen Ebenen auf ihre Vollstndigkeit berprfen
     For y:=0 To MaxHeight Do Begin
         Complete[y]:=True;
         For x:=0 To MaxWidth Do Begin
             Ok:=True;
             For z:=0 To MaxWidth Do Begin
                 If aField[x, z, y] = 0 Then Begin
                    Complete[y]:=False;
                    Ok:=False;
                    Break;
                 End;
             End;
             If Not Ok Then Break;
         End;
     End;
     // Alle vollstndigen Ebenen entfernen
     i:=0;
     FillChar(Tmp, SizeOf(Tmp), #0);
     For y:=0 To MaxHeight Do Begin
         If Not Complete[y] Then Begin
            For x:=0 To MaxWidth Do For z:=0 To MaxWidth Do Tmp[x, z, i]:=aField[x, z, y];
            Inc(i);
         End;
     End;
     Move(Tmp, aField, SizeOf(aField));
End;

Procedure TFrmTetris.InitNewBlock;
Begin
     iActBlock:=iNextBlock;
     iNextBlock:=Random(MaxBlock+1);
     BlockPreview:=aBlocks[iNextBlock];
     With BlockPos Do Begin
          x:=(MaxWidth+1) Div 2;
          y:=MaxHeight+aBlocks[iActBlock].Height-1;
          z:=(MaxWidth+1) Div 2;
          RotX:=0;
          RotY:=0;
          RotZ:=Pi;
     End;
End;

Procedure TFrmTetris.MoveBlockDown;
Begin
     If Not TestBlock(BlockPos.x, BlockPos.y-1, BlockPos.z, BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ) Then Begin
        InsertBlock;

        Case iGameMode Of
             gmLines: RemoveCompleteLines;
             gmPlanes: RemoveCompletePlanes;
        End;

        InitNewBlock;
        TestBlock(BlockPos.x, BlockPos.y, BlockPos.z, BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ);
     End;
End;

Procedure TFrmTetris.DrawGame;
Var
   sx, sy, sz, cx, cy, cz: Double;
   i, BlockTexIndex, x, y, z: Integer;
   x2, y2, z2: Integer;
   a, b, c, a2, b2, c2: Double;
   ColorFading: Double;
   aGrid: Array[1..4] Of Boolean;
   Degree: Integer;
   Keyboardstate: TKeyboardstate;
   TimeBeforeDrawing: Cardinal;
   Tmp, Tmp2: TBlockColor;
Begin
     // Timer starten
     timeBeginPeriod(1);
     TimeBeforeDrawing:=timeGetTime;

     // Winkel von Bogenma in Grad umrechnen
     Degree:=Round(sgRotY*(180/Pi));

     // Gitter je nach Winkel ein/ausschalten
     Case Degree of
            0.. 89:Begin
                        aGrid[1]:=True;
                        aGrid[2]:=True;
                        aGrid[3]:=False;
                        aGrid[4]:=False;
                   End;
           90..179:Begin
                        aGrid[1]:=True;
                        aGrid[2]:=False;
                        aGrid[3]:=False;
                        aGrid[4]:=True;
                   End;
          180..269:Begin
                        aGrid[1]:=False;
                        aGrid[2]:=False;
                        aGrid[3]:=True;
                        aGrid[4]:=True;
                   End;
          270..359:Begin
                        aGrid[1]:=False;
                        aGrid[2]:=True;
                        aGrid[3]:=True;
                        aGrid[4]:=False;
                   End;
     End;

     // Die Tastaturbelegung an den Winkel anpassen
     Case Degree Of
          316..360, 0..44: Begin
                        PosIncFac.x:=1;
                        PosIncFac.z:=1;
                        PosIncFac.LeftRight:=LeftRightChangeXAxis;
                   End;
          45..134: Begin
                        PosIncFac.x:=1;
                        PosIncFac.z:=-1;
                        PosIncFac.LeftRight:=LeftRightChangeZAxis;
                   End;
          135..224: Begin
                        PosIncFac.x:=-1;
                        PosIncFac.z:=-1;
                        PosIncFac.LeftRight:=LeftRightChangeXAxis;
                   End;
          225..315: Begin
                        PosIncFac.x:=-1;
                        PosIncFac.z:=1;
                        PosIncFac.LeftRight:=LeftRightChangeZAxis;
                   End;
     End;

     // Sinus- und Cosinus-Werte berechnen (fr Rotation)
     sx:=sin(sgRotX);
     sy:=sin(sgRotY);
     sz:=sin(sgRotZ);
     cx:=cos(sgRotX);
     cy:=cos(sgRotY);
     cz:=cos(sgRotZ);

     // Eckpunkte des Gitternetzes erstellen
     iVerteciesNum:=0;
     If aGrid[1] Then Begin
        For z:=-((MaxWidth+1) Div 2 ) To ((MaxWidth+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            aVertecies[iVerteciesNum].y:=-(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum].z:=z * BlockSize;
            aVertecies[iVerteciesNum+1].x:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            aVertecies[iVerteciesNum+1].y:=(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].z:=z * BlockSize;
            Inc(iVerteciesNum,2);
        End;
        For y:=-((MaxHeight+1) Div 2 ) To ((MaxHeight+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            aVertecies[iVerteciesNum].y:=y * BlockSize;
            aVertecies[iVerteciesNum].z:=-(MaxWidth+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].x:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            aVertecies[iVerteciesNum+1].y:=y * BlockSize;
            aVertecies[iVerteciesNum+1].z:=(MaxWidth+1) Div 2 * BlockSize;
            Inc(iVerteciesNum,2);
        End;
     End;
     If aGrid[2] Then Begin
        For x:=-((MaxWidth+1) Div 2 ) To ((MaxWidth+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=x * BlockSize;
            aVertecies[iVerteciesNum].y:=-(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum].z:=(MaxWidth+1) Div 2 * BlockSize + 1;
            aVertecies[iVerteciesNum+1].x:=x * BlockSize;
            aVertecies[iVerteciesNum+1].y:=(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].z:=(MaxWidth+1) Div 2 * BlockSize + 1;
            Inc(iVerteciesNum,2);
        End;
        For y:=-((MaxHeight+1) Div 2 ) To ((MaxHeight+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=-(MaxWidth+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum].y:=y * BlockSize;
            aVertecies[iVerteciesNum].z:=(MaxWidth+1) Div 2 * BlockSize + 1;
            aVertecies[iVerteciesNum+1].x:=(MaxWidth+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].y:=y * BlockSize;
            aVertecies[iVerteciesNum+1].z:=(MaxWidth+1) Div 2 * BlockSize + 1;
            Inc(iVerteciesNum,2);
        End;
     End;
     If aGrid[3] Then Begin
        For z:=-((MaxWidth+1) Div 2 ) To ((MaxWidth+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=(MaxWidth+1) Div 2 * BlockSize + 1;
            aVertecies[iVerteciesNum].y:=-(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum].z:=z * BlockSize;
            aVertecies[iVerteciesNum+1].x:=(MaxWidth+1) Div 2 * BlockSize + 1;
            aVertecies[iVerteciesNum+1].y:=(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].z:=z * BlockSize;
            Inc(iVerteciesNum,2);
        End;
        For y:=-((MaxHeight+1) Div 2 ) To ((MaxHeight+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=(MaxWidth+1) Div 2 * BlockSize + 1;
            aVertecies[iVerteciesNum].y:=y * BlockSize;
            aVertecies[iVerteciesNum].z:=-(MaxWidth+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].x:=(MaxWidth+1) Div 2 * BlockSize + 1;
            aVertecies[iVerteciesNum+1].y:=y * BlockSize;
            aVertecies[iVerteciesNum+1].z:=(MaxWidth+1) Div 2 * BlockSize;
            Inc(iVerteciesNum,2);
        End;
     End;
     If aGrid[4] Then Begin
        For x:=-((MaxWidth+1) Div 2 ) To ((MaxWidth+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=x * BlockSize;
            aVertecies[iVerteciesNum].y:=-(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum].z:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            aVertecies[iVerteciesNum+1].x:=x * BlockSize;
            aVertecies[iVerteciesNum+1].y:=(MaxHeight+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].z:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            Inc(iVerteciesNum,2);
        End;
        For y:=-((MaxHeight+1) Div 2 ) To ((MaxHeight+1) Div 2) Do Begin
            aVertecies[iVerteciesNum].x:=-(MaxWidth+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum].y:=y * BlockSize;
            aVertecies[iVerteciesNum].z:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            aVertecies[iVerteciesNum+1].x:=(MaxWidth+1) Div 2 * BlockSize;
            aVertecies[iVerteciesNum+1].y:=y * BlockSize;
            aVertecies[iVerteciesNum+1].z:=-(MaxWidth+1) Div 2 * BlockSize - 1;
            Inc(iVerteciesNum,2);
        End;
     End;
     iGridVerteciesNum:=iVerteciesNum;
     For i:=0 To iGridVerteciesNum-1 Do Begin
         aVertecies[i].r:=255;
         aVertecies[i].g:=255;
         aVertecies[i].b:=255;
     End;

     // Eckpunkte fr Schatten erstellen
     CreateShadow;

     // Eckpunkte der Trgerplatte erstellen
     x:=-((MaxWidth+1)*BlockSize) Div 2 - BlockSize - (BlockSize Div 2);
     y:=((MaxHeight+1)*BlockSize) Div 2 + BlockSize;
     z:=-((MaxWidth+1)*BlockSize) Div 2 - BlockSize - (BlockSize Div 2);

     aVertecies[iVerteciesNum].x:=x;
     aVertecies[iVerteciesNum].y:=y;
     aVertecies[iVerteciesNum].z:=z;
     aVertecies[iVerteciesNum].tmuvtx[0].sow:=0;
     aVertecies[iVerteciesNum].tmuvtx[0].tow:=255;

     aVertecies[iVerteciesNum+1].x:=x;
     aVertecies[iVerteciesNum+1].y:=y;
     aVertecies[iVerteciesNum+1].z:=-z;
     aVertecies[iVerteciesNum+1].tmuvtx[0].sow:=0;
     aVertecies[iVerteciesNum+1].tmuvtx[0].tow:=0;

     aVertecies[iVerteciesNum+2].x:=-x;
     aVertecies[iVerteciesNum+2].y:=y;
     aVertecies[iVerteciesNum+2].z:=-z;
     aVertecies[iVerteciesNum+2].tmuvtx[0].sow:=255;
     aVertecies[iVerteciesNum+2].tmuvtx[0].tow:=0;

     aVertecies[iVerteciesNum+3].x:=-x;
     aVertecies[iVerteciesNum+3].y:=y;
     aVertecies[iVerteciesNum+3].z:=z;
     aVertecies[iVerteciesNum+3].tmuvtx[0].sow:=255;
     aVertecies[iVerteciesNum+3].tmuvtx[0].tow:=255;

     aVertecies[iVerteciesNum+4].x:=x;
     aVertecies[iVerteciesNum+4].y:=y-BlockSize;
     aVertecies[iVerteciesNum+4].z:=z;
     aVertecies[iVerteciesNum+4].tmuvtx[0].sow:=25;
     aVertecies[iVerteciesNum+4].tmuvtx[0].tow:=231;

     aVertecies[iVerteciesNum+5].x:=x;
     aVertecies[iVerteciesNum+5].y:=y-BlockSize;
     aVertecies[iVerteciesNum+5].z:=-z;
     aVertecies[iVerteciesNum+5].tmuvtx[0].sow:=25;
     aVertecies[iVerteciesNum+5].tmuvtx[0].tow:=25;

     aVertecies[iVerteciesNum+6].x:=-x;
     aVertecies[iVerteciesNum+6].y:=y-BlockSize;
     aVertecies[iVerteciesNum+6].z:=-z;
     aVertecies[iVerteciesNum+6].tmuvtx[0].sow:=231;
     aVertecies[iVerteciesNum+6].tmuvtx[0].tow:=25;

     aVertecies[iVerteciesNum+7].x:=-x;
     aVertecies[iVerteciesNum+7].y:=y-BlockSize;
     aVertecies[iVerteciesNum+7].z:=z;
     aVertecies[iVerteciesNum+7].tmuvtx[0].sow:=231;
     aVertecies[iVerteciesNum+7].tmuvtx[0].tow:=231;

     For i:=0 To 3 Do Begin
         aVertecies[iVerteciesNum+i].r:=0;
         aVertecies[iVerteciesNum+i].g:=0;
         aVertecies[iVerteciesNum+i].b:=0;
     End;
     For i:=4 To 7 Do Begin
         aVertecies[iVerteciesNum+i].r:=255;
         aVertecies[iVerteciesNum+i].g:=255;
         aVertecies[iVerteciesNum+i].b:=255;
     End;

     // Stein in die Matrix einfgen
     InsertBlock;

     // Eckpunkte der in der Matrix enthaltenen Blcke erstellen
     Inc(iVerteciesNum,8);
     iTexIndeciesNum:=0;
     x2:=-((MaxWidth+1)*BlockSize) Div 2;
     For x:=0 To MaxWidth Do Begin

         z2:=((MaxWidth+1)*BlockSize) Div 2 - BlockSize;
         For z:=0 To MaxWidth Do Begin

             y2:=((MaxHeight+1)*BlockSize) Div 2;
             For y:=0 To MaxHeight Do Begin

                 If aField[x, z, y] > 0 Then Begin

                    If Not Enclosed(x, y, z) Then Begin

                       aTexIndecies[iTexIndeciesNum]:=aField[x, z, y]+BlockTexOffset;
                       Inc(iTexIndeciesNum);

                       Move(aBlockColors[aField[x, z, y]-1].r, Tmp, 12);
                       Tmp2.r:=Round(Tmp.r*0.7);
                       Tmp2.g:=Round(Tmp.g*0.7);
                       Tmp2.b:=Round(Tmp.b*0.7);

                       aVertecies[iVerteciesNum].x:=x2;
                       aVertecies[iVerteciesNum].y:=y2;
                       aVertecies[iVerteciesNum].z:=z2;
                       Move(Tmp2, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2;
                       aVertecies[iVerteciesNum].y:=y2;
                       aVertecies[iVerteciesNum].z:=z2+BlockSize;;
                       Move(Tmp2, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2+BlockSize;
                       aVertecies[iVerteciesNum].y:=y2;
                       aVertecies[iVerteciesNum].z:=z2+BlockSize;
                       Move(Tmp2, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2+BlockSize;
                       aVertecies[iVerteciesNum].y:=y2;
                       aVertecies[iVerteciesNum].z:=z2;
                       Move(Tmp2, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2;
                       aVertecies[iVerteciesNum].y:=y2-BlockSize;
                       aVertecies[iVerteciesNum].z:=z2;
                       Move(Tmp, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2;
                       aVertecies[iVerteciesNum].y:=y2-BlockSize;
                       aVertecies[iVerteciesNum].z:=z2+BlockSize;
                       Move(Tmp, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2+BlockSize;
                       aVertecies[iVerteciesNum].y:=y2-BlockSize;
                       aVertecies[iVerteciesNum].z:=z2+BlockSize;
                       Move(Tmp, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                       aVertecies[iVerteciesNum].x:=x2+BlockSize;
                       aVertecies[iVerteciesNum].y:=y2-BlockSize;
                       aVertecies[iVerteciesNum].z:=z2;
                       Move(Tmp, aVertecies[iVerteciesNum].r, 12);
                       Inc(iVerteciesNum);

                    End;

                 End;
                 Dec(y2, BlockSize);

             End;
             Dec(z2, BlockSize);

         End;
         Inc(x2, BlockSize);

     End;

     // Stein entfernen
     RemoveBlock;

     // Vertecies transformieren und translieren
     For i:=0 To iVerteciesNum-1 Do Begin

         // Drehung um Y-Achse
         a:=(aVertecies[i].x*cy)+(aVertecies[i].z*sy);
         b:=aVertecies[i].y;
         c:=-(aVertecies[i].x*sy)+(aVertecies[i].z*cy);

         // Drehung um X-Achse
         a2:=a;
         b2:=(b*cx)-(c*sx);
         c2:=(b*sx)+(c*cx);

         // Object-Space -> Screen-Space
         aVertecies[i].z:=c2+iZoom; // Nach hinten verschieben
         aVertecies[i].x:=Round((EyeDistance * a2) / aVertecies[i].z)+512;
         aVertecies[i].y:=Round((EyeDistance * b2) / aVertecies[i].z)+384;

         // 1 / Z fr W-Buffer...
         aVertecies[i].oow:=1 / aVertecies[i].z;

         // Farbverlauf
         ColorFading:=800 / aVertecies[i].z - 1;
         ColorFading:=Sqr(ColorFading);
         If ColorFading > 1 Then ColorFading:=1;
         aVertecies[i].r:=Round(aVertecies[i].r * ColorFading);
         aVertecies[i].g:=Round(aVertecies[i].g * ColorFading);
         aVertecies[i].b:=Round(aVertecies[i].b * ColorFading);

     End;

     // Bild zeichnen
     GrAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
     grBufferClear(0, 0, GR_WDEPTHVALUE_FARTHEST);

     // Sterne zeichnen
     DrawStars;
     MoveStars;

     // Gitternetz(e) zeichnen
     guColorCombineFunction(GR_COLORCOMBINE_ITRGB);
     i:=0;
     While i < iGridVerteciesNum Do Begin
           grDrawLine(aVertecies[i], aVertecies[i+1]);
           Inc(i, 2);
     End;

     // Texturierte Trgerplatte zeichnen
     guColorCombineFunction(GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB);
     grTexClampMode(GR_TMU0, GR_TEXTURECLAMP_WRAP, GR_TEXTURECLAMP_WRAP);
     grTexSource(GR_TMU0, aTextureAddresses[0], GR_MIPMAPLEVELMASK_BOTH, aTextures[0]);
     grTexCombine(GR_TMU0, GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE,
                  GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE, FXFALSE, FXFALSE);
     For i:=iGridVerteciesNum+iShadowVerteciesNum To iGridVerteciesNum+iShadowVerteciesNum+8 Do Begin
         aVertecies[i].tmuvtx[0].sow:=aVertecies[i].tmuvtx[0].sow / aVertecies[i].z;
         aVertecies[i].tmuvtx[0].tow:=aVertecies[i].tmuvtx[0].tow / aVertecies[i].z;
     End;
     i:=iGridVerteciesNum+iShadowVerteciesNum;
     // Unten
{     grDrawTriangle(aVertecies[i], aVertecies[i+1], aVertecies[i+2]);
     grDrawTriangle(aVertecies[i+2], aVertecies[i+3], aVertecies[i]);}

     // Oben
     grDrawTriangle(aVertecies[i+4], aVertecies[i+7], aVertecies[i+6]);
     grDrawTriangle(aVertecies[i+6], aVertecies[i+5], aVertecies[i+4]);

     // Vorne
     grDrawTriangle(aVertecies[i+4], aVertecies[i], aVertecies[i+3]);
     grDrawTriangle(aVertecies[i+3], aVertecies[i+7], aVertecies[i+4]);

     // Links
     grDrawTriangle(aVertecies[i+5], aVertecies[i+1], aVertecies[i]);
     grDrawTriangle(aVertecies[i+5], aVertecies[i], aVertecies[i+4]);

     // Hinten
     grDrawTriangle(aVertecies[i+6], aVertecies[i+2], aVertecies[i+1]);
     grDrawTriangle(aVertecies[i+6], aVertecies[i+1], aVertecies[i+5]);

     // Rechts
     grDrawTriangle(aVertecies[i+7], aVertecies[i+3], aVertecies[i+2]);
     grDrawTriangle(aVertecies[i+7], aVertecies[i+2], aVertecies[i+6]);

     // Spielsteine rsche malen
     i:=iGridVerteciesNum+iShadowVerteciesNum+8;
     BlockTexIndex:=0;
     While i < iVerteciesNum Do Begin

           // Textur whlen
           grTexSource(GR_TMU0, aTextureAddresses[aTexIndecies[BlockTexIndex]],
                       GR_MIPMAPLEVELMASK_BOTH, aTextures[aTexIndecies[BlockTexIndex]]);

           // Unten
           aVertecies[i].tmuvtx[0].sow:=0;
           aVertecies[i].tmuvtx[0].tow:=0;
           aVertecies[i+1].tmuvtx[0].sow:=0;
           aVertecies[i+1].tmuvtx[0].tow:=255 / aVertecies[i+1].z;
           aVertecies[i+2].tmuvtx[0].sow:=255 / aVertecies[i+2].z;
           aVertecies[i+2].tmuvtx[0].tow:=aVertecies[i+2].tmuvtx[0].sow;
           aVertecies[i+3].tmuvtx[0].sow:=255 / aVertecies[i+3].z;
           aVertecies[i+3].tmuvtx[0].tow:=0;
           grDrawTriangle(aVertecies[i], aVertecies[i+1], aVertecies[i+2]);
           grDrawTriangle(aVertecies[i+2], aVertecies[i+3], aVertecies[i]);

           // Oben
           aVertecies[i+4].tmuvtx[0].sow:=0;
           aVertecies[i+4].tmuvtx[0].tow:=255 / aVertecies[i+4].z;
           aVertecies[i+5].tmuvtx[0].sow:=0;
           aVertecies[i+5].tmuvtx[0].tow:=0;
           aVertecies[i+6].tmuvtx[0].sow:=255 / aVertecies[i+6].z;
           aVertecies[i+6].tmuvtx[0].tow:=0;
           aVertecies[i+7].tmuvtx[0].sow:=255 / aVertecies[i+7].z;
           aVertecies[i+7].tmuvtx[0].tow:=aVertecies[i+7].tmuvtx[0].sow;
           grDrawTriangle(aVertecies[i+4], aVertecies[i+7], aVertecies[i+6]);
           grDrawTriangle(aVertecies[i+6], aVertecies[i+5], aVertecies[i+4]);

           // Vorne
           aVertecies[i].tmuvtx[0].sow:=0;
           aVertecies[i].tmuvtx[0].tow:=255 / aVertecies[i+4].z;
           aVertecies[i+3].tmuvtx[0].sow:=255 / aVertecies[i+3].z;
           aVertecies[i+3].tmuvtx[0].tow:=aVertecies[i+3].tmuvtx[0].sow;
           aVertecies[i+4].tmuvtx[0].sow:=0;
           aVertecies[i+4].tmuvtx[0].tow:=0;
           aVertecies[i+7].tmuvtx[0].sow:=255 / aVertecies[i+7].z;
           aVertecies[i+7].tmuvtx[0].tow:=0;
           grDrawTriangle(aVertecies[i+4], aVertecies[i], aVertecies[i+3]);
           grDrawTriangle(aVertecies[i+3], aVertecies[i+7], aVertecies[i+4]);

           // Links
           aVertecies[i].tmuvtx[0].sow:=255 / aVertecies[i].z;
           aVertecies[i].tmuvtx[0].tow:=aVertecies[i].tmuvtx[0].sow;
           aVertecies[i+1].tmuvtx[0].sow:=0;
           aVertecies[i+1].tmuvtx[0].tow:=255 / aVertecies[i+1].z;
           aVertecies[i+4].tmuvtx[0].sow:=255 / aVertecies[i+4].z;
           aVertecies[i+4].tmuvtx[0].tow:=0;
           aVertecies[i+5].tmuvtx[0].sow:=0;
           aVertecies[i+5].tmuvtx[0].tow:=0;
           grDrawTriangle(aVertecies[i+5], aVertecies[i+1], aVertecies[i]);
           grDrawTriangle(aVertecies[i+5], aVertecies[i], aVertecies[i+4]);

           // Hinten
           aVertecies[i+1].tmuvtx[0].sow:=255 / aVertecies[i+1].z;
           aVertecies[i+1].tmuvtx[0].tow:=aVertecies[i+1].tmuvtx[0].sow;
           aVertecies[i+2].tmuvtx[0].sow:=0;
           aVertecies[i+2].tmuvtx[0].tow:=255 / aVertecies[i+2].z;
           aVertecies[i+5].tmuvtx[0].sow:=255 / aVertecies[i+5].z;
           aVertecies[i+5].tmuvtx[0].tow:=0;
           aVertecies[i+6].tmuvtx[0].sow:=0;
           aVertecies[i+6].tmuvtx[0].tow:=0;
           grDrawTriangle(aVertecies[i+6], aVertecies[i+2], aVertecies[i+1]);
           grDrawTriangle(aVertecies[i+6], aVertecies[i+1], aVertecies[i+5]);

           // Rechts
           aVertecies[i+2].tmuvtx[0].sow:=255 / aVertecies[i+2].z;
           aVertecies[i+2].tmuvtx[0].tow:=aVertecies[i+2].tmuvtx[0].sow;
           aVertecies[i+3].tmuvtx[0].sow:=0;
           aVertecies[i+3].tmuvtx[0].tow:=255 / aVertecies[i+3].z;
           aVertecies[i+6].tmuvtx[0].sow:=255 / aVertecies[i+6].z;
           aVertecies[i+6].tmuvtx[0].tow:=0;
           aVertecies[i+7].tmuvtx[0].sow:=0;
           aVertecies[i+7].tmuvtx[0].tow:=0;
           grDrawTriangle(aVertecies[i+7], aVertecies[i+3], aVertecies[i+2]);
           grDrawTriangle(aVertecies[i+7], aVertecies[i+2], aVertecies[i+6]);

           // Nchster Quader/Wrfel
           Inc(i, 8);
           Inc(BlockTexIndex);
     End;

     // Schatten zeichnen
     grConstantColorValue(192); // 75% der Ursprungsfarbe
     grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_OTHER, GR_COMBINE_FACTOR_ONE, GR_COMBINE_LOCAL_NONE, GR_COMBINE_OTHER_CONSTANT, FXFALSE);
     grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
     i:=iGridVerteciesNum;
     While i-iGridVerteciesNum < iShadowVerteciesNum Do Begin
           grDrawTriangle(aVertecies[i], aVertecies[i+1], aVertecies[i+2]);
           grDrawTriangle(aVertecies[i+2], aVertecies[i+3], aVertecies[i]);
           Inc(i, 4);
     End;

     // Vorschau auf nchsten Stein zeichnen
     DrawPreviewBlockBackground;
     DrawPreviewBlock;

     // Text zeichnen
     If LFB_Lock <> Nil Then
        OutTextXY(0, 0, 0, 5, Stri(Round(fps)));
     LFB_UnLock;

     // Berechneten Frame anzeigen
     grBufferSwap(0);

     // Frame-Zeit stoppen
     iFrameTime:=timeGetTime-TimeBeforeDrawing;
     timeEndPeriod(1);
     // Fr den unwahrscheinlichen Fall, da der Rechner mehr als 1000 f/s schafft...
     If iFrameTime <= 0 Then iFrameTime:=1;
     fps:=1000/iFrameTime;
     iStarsSpeed:=Round(IncFacPerMS.Stars*iFrameTime);

     // Keyboard-Abfrage
     GetKeyboardState(Keyboardstate);
     If (Keyboardstate[VK_MENU] And 128) > 0 Then Begin
        If (Keyboardstate[VK_LEFT] And 128) > 0 Then Begin
           sgRotY:=sgRotY+(IncFacPerMS.RotY*iFrameTime);
           If sgRotY >= Pi*2 Then sgRotY:=sgRotY-(Pi*2);
        End;
        If (Keyboardstate[VK_RIGHT] And 128) > 0 Then Begin
           sgRotY:=sgRotY-(IncFacPerMS.RotY*iFrameTime);
           If sgRotY < 0 Then sgRotY:=(Pi*2)+sgRotY;
        End;
        If (Keyboardstate[VK_UP] And 128) > 0 Then Begin
           sgRotX:=sgRotX+(IncFacPerMS.RotX*iFrameTime);
           If sgRotX >= Pi/2 Then sgRotX:=Pi/2;
        End;
        If (Keyboardstate[VK_DOWN] And 128) > 0 Then Begin
           sgRotX:=sgRotX-(IncFacPerMS.RotX*iFrameTime);
           If sgRotX < 0 Then sgRotX:=0;
        End;
     End;

     // Ticker
     If (KeyboardState[VK_MENU] And 128) = 0 Then Begin
        sgTickCounter:=sgTickCounter+(IncFacPerMS.GameSpeed*iFrameTime);
        If sgTickCounter >= iGameSpeed Then Begin
           sgTickCounter:=0;
           MoveBlockDown;
        End;
     End;

End;

Procedure TFrmTetris.CreateShadow;
Var
   x, y, z: Integer;
   a, b, c: Integer;
   TmpPlane: Array[0..4, 0..4] Of ShortInt;
   TmpHeight: Array[0..4, 0..4] Of Byte;

Begin
     FillChar(TmpHeight, Sizeof(TmpPlane), #0);
     For x:=0 To 4 Do Begin
         For z:=0 To 4 Do Begin
             TmpPlane[x, z]:=-1;
             For y:=0 To 4 Do Begin
                 If CurrentBlock.Block[x, y, z] > 0 Then Begin
                    TmpPlane[x, z]:=y;
                    Break;
                 End;
             End;
         End;
     End;
     For x:=0 To 4 Do Begin
         For z:=0 To 4 Do Begin
             b:=BlockPos.y-TmpPlane[x, z];
             If b > MaxHeight Then b:=MaxHeight;
             If b >= 0 Then Begin
                For y:=0 To b Do Begin
                    If (x+BlockPos.x-2 >= 0) And (x+BlockPos.x-2 <= MaxWidth) And
                       (z+BlockPos.z-2 >= 0) And (z+BlockPos.z-2 <= MaxWidth) Then
                          If (aField[x+BlockPos.x-2, z+BlockPos.z-2, y] > 0) And (TmpPlane[x, z] >= 0) Then
                             TmpHeight[x, z]:=y+1;
                End;
             End;
         End;
     End;

     iShadowVerteciesNum:=0;
     a:=-((MaxWidth+1)*BlockSize) Div 2 + (BlockPos.x-2) * BlockSize;
     For x:=0 To 4 Do Begin
         c:=((MaxWidth+1)*BlockSize) Div 2 - BlockSize - (BlockPos.z-2) * BlockSize;
         For z:=0 To 4 Do Begin
             If (TmpPlane[x, z] >= 0) Then Begin

                b:=(((MaxHeight+1)*BlockSize) Div 2) - (TmpHeight[x, z] * BlockSize) - 1;

                aVertecies[iVerteciesNum].x:=a;
                aVertecies[iVerteciesNum].y:=b;
                aVertecies[iVerteciesNum].z:=c;
                aVertecies[iVerteciesNum].r:=0;
                aVertecies[iVerteciesNum].g:=0;
                aVertecies[iVerteciesNum].b:=0;
                Inc(iVerteciesNum);

                aVertecies[iVerteciesNum].x:=a+BlockSize;
                aVertecies[iVerteciesNum].y:=b;
                aVertecies[iVerteciesNum].z:=c;
                aVertecies[iVerteciesNum].r:=0;
                aVertecies[iVerteciesNum].g:=0;
                aVertecies[iVerteciesNum].b:=0;
                Inc(iVerteciesNum);

                aVertecies[iVerteciesNum].x:=a+BlockSize;
                aVertecies[iVerteciesNum].y:=b;
                aVertecies[iVerteciesNum].z:=c+BlockSize;
                aVertecies[iVerteciesNum].r:=0;
                aVertecies[iVerteciesNum].g:=0;
                aVertecies[iVerteciesNum].b:=0;
                Inc(iVerteciesNum);

                aVertecies[iVerteciesNum].x:=a;
                aVertecies[iVerteciesNum].y:=b;
                aVertecies[iVerteciesNum].z:=c+BlockSize;
                aVertecies[iVerteciesNum].r:=0;
                aVertecies[iVerteciesNum].g:=0;
                aVertecies[iVerteciesNum].b:=0;
                Inc(iVerteciesNum);

                Inc(iShadowVerteciesNum, 4);
             End;
             Dec(c, BlockSize);
         End;
         Inc(a, BlockSize);
     End;
End;

procedure TFrmTetris.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
     Case iGameState Of
          gsGame:
             Begin
                  If (GetKeyState(VK_MENU) And 128) = 0 Then
                  Case Key Of
                       VK_LEFT:
                          Begin
                               If PosIncFac.LeftRight = LeftRightChangeXAxis Then
                                  TestBlock(BlockPos.x-PosIncFac.x, BlockPos.y, BlockPos.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ)
                               Else
                                  TestBlock(BlockPos.x, BlockPos.y, BlockPos.z-PosIncFac.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ);
                          End;
                       VK_RIGHT:
                          Begin
                               If PosIncFac.LeftRight = LeftRightChangeXAxis Then
                                  TestBlock(BlockPos.x+PosIncFac.x, BlockPos.y, BlockPos.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ)
                               Else
                                   TestBlock(BlockPos.x, BlockPos.y, BlockPos.z+PosIncFac.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ);
                          End;
                       VK_UP:
                          Begin
                               If PosIncFac.LeftRight = LeftRightChangeXAxis Then
                                  TestBlock(BlockPos.x, BlockPos.y, BlockPos.z-PosIncFac.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ)
                               Else
                                  TestBlock(BlockPos.x-PosIncFac.x, BlockPos.y, BlockPos.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ);
                          End;
                       VK_DOWN:
                          Begin
                               If PosIncFac.LeftRight = LeftRightChangeXAxis Then
                                  TestBlock(BlockPos.x, BlockPos.y, BlockPos.z+PosIncFac.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ)
                               Else
                                  TestBlock(BlockPos.x+PosIncFac.x, BlockPos.y, BlockPos.z,
                                            BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ);
                          End;
                       VK_SPACE: MoveBlockDown;
                       VK_ESCAPE: Close;
                       Else Begin
                          If Key = Word(VKKeyScan('y')) Then TestBlock(BlockPos.x, BlockPos.y, BlockPos.z,
                                                                 BlockPos.RotX+pi/2, BlockPos.RotY, BlockPos.RotZ);
                          If Key = Word(VKKeyScan('x')) Then TestBlock(BlockPos.x, BlockPos.y, BlockPos.z,
                                                                 BlockPos.RotX, BlockPos.RotY+pi/2, BlockPos.RotZ);
                          If Key = Word(VKKeyScan('c')) Then TestBlock(BlockPos.x, BlockPos.y, BlockPos.z,
                                                                 BlockPos.RotX, BlockPos.RotY, BlockPos.RotZ+pi/2);
                       End;
                  End;
             End;
          gsMainMenu:
             Begin
                  Case Key Of
                       VK_Escape: Close;
                       VK_UP:
                          Begin
                               iRotMenu:=1;
                          End;
                       VK_DOWN:
                          Begin
                               iRotMenu:=-1;
                          End;
                       Else iGameState:=gsGame;
                  End;
             End;
     End;
End;

Procedure TFrmTetris.InitFont;
Var
   i: Integer;
   FontSprite: LFB_Sprite;
Begin
     aFontFiles[FontNr]:=FontFile;
     LFB_LoadPCXFile(aFontFiles[FontNr], FontSprite);

     For i:=0 To 255 Do Begin
         If aFontCoordinates[FontNr, i].Width > 0 Then
            LFB_GetSpriteFromSprite(FontSprite, aFonts[FontNr, i],
               aFontCoordinates[FontNr, i].x, aFontCoordinates[FontNr, i].y,
               aFontCoordinates[FontNr, i].Width, aFontCoordinates[FontNr, i].Height);
     End;

     LFB_FreeSprite(FontSprite);
End;

Procedure TFrmTetris.OutTextXY;
Var
   i: Integer;
Begin
     For i:=1 To Length(Text) Do Begin
         If aFonts[FontNr, Ord(Text[i])].Width > 0 Then Begin
            LFB_PutSprite(x, y, aFonts[FontNr, Ord(Text[i])]);
            Inc(x, aFontCoordinates[FontNr, Ord(Text[i])].Width + FontSpacing);
         End Else Inc(x, FontSpacing);
     End;
End;

Procedure TFrmTetris.InitPreviewBlockBackground;
Var
   i: Integer;
   Angle: Single;
   AngelInc: Single;
Begin
     aPreviewVertecies[0].x:=PreviewBlockCenterX;
     aPreviewVertecies[0].y:=PreviewBlockCenterY;
     aPreviewVertecies[0].a:=125;
     Angle:=0;
     AngelInc:=(2*Pi)/PreviewBlockPolygons;
     For i:=1 To PreviewBlockPolygons Do Begin
         aPreviewVertecies[i].x:=Round(-PreviewBlockRadius*Sin(Angle))+PreviewBlockCenterX;
         aPreviewVertecies[i].y:=Round(PreviewBlockRadius*Cos(Angle))+PreviewBlockCenterY;
         aPreviewVertecies[i].a:=0;
         Angle:=Angle+AngelInc;
     End;
     For i:=0 To PreviewBlockPolygons Do Begin
         aPreviewVertecies[i].r:=255;
         aPreviewVertecies[i].g:=0;
         aPreviewVertecies[i].b:=125;
         aPreviewVertecies[i].oow:=1/1000;
     End;
End;

Procedure TFrmTetris.DrawPreviewBlockBackground;
Var
   i: Integer;
Begin
     grCullMode(GR_CULL_DISABLE);
     guColorCombineFunction(GR_COLORCOMBINE_ITRGB);
     grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_LOCAL, GR_COMBINE_FACTOR_ZERO, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_NONE, FXFALSE);
     grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
     For i:=1 To PreviewBlockPolygons-1 Do Begin
         grDrawTriangle(aPreviewVertecies[0], aPreviewVertecies[i], aPreviewVertecies[i+1]);
     End;
     grDrawTriangle(aPreviewVertecies[0], aPreviewVertecies[PreviewBlockPolygons], aPreviewVertecies[1]);
     grCullMode(GR_CULL_POSITIVE);
End;

Procedure TFrmTetris.DrawPreviewBlock;
Var
   Vertecies: Array[0..999] Of TgrVertex;
   TexIndecies: Array[0..1499] Of Integer;
   VerteciesNum: Integer;
   TexIndeciesNum: Integer;
   sx, sy, sz, cx, cy, cz: Double;
   i, BlockTexIndex: Integer;
   x, y, z: Integer;
   x2, y2, z2: Integer;
   a, b, c, a2, b2, c2: Double;
   Tmp, Tmp2: TBlockColor;
Begin
     // Sinus- und Cosinus-Werte berechnen (fr Rotation)
     sx:=sin(sgRotXp);
     sy:=sin(sgRotYp);
     sz:=sin(sgRotZp);
     cx:=cos(sgRotXp);
     cy:=cos(sgRotYp);
     cz:=cos(sgRotZp);

     // Eckpunkte der in der Matrix enthaltenen Blcke erstellen
     VerteciesNum:=0;
     TexIndeciesNum:=0;
     x2:=-(Integer(BlockPreview.Width)*BlockSize) Div 2;
     For x:=0 To Integer(BlockPreview.Width)-1 Do Begin

         z2:=(Integer(BlockPreview.Depth)*BlockSize) Div 2 - BlockSize;
         For z:=0 To Integer(BlockPreview.Depth)-1 Do Begin

             y2:=(Integer(BlockPreview.Height)*BlockSize) Div 2;
             For y:=0 To Integer(BlockPreview.Height)-1 Do Begin

                 If BlockPreview.Block[x, y, z] > 0 Then Begin

                    TexIndecies[TexIndeciesNum]:=BlockPreview.Block[x, y, z]+BlockTexOffset;
                    Inc(TexIndeciesNum);

                    Move(aBlockColors[BlockPreview.Block[x, y, z]-1].r, Tmp, 12);
                    Tmp2.r:=Round(Tmp.r*0.7);
                    Tmp2.g:=Round(Tmp.g*0.7);
                    Tmp2.b:=Round(Tmp.b*0.7);

                    Vertecies[VerteciesNum].x:=x2;
                    Vertecies[VerteciesNum].y:=y2;
                    Vertecies[VerteciesNum].z:=z2;
                    Move(Tmp2, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2;
                    Vertecies[VerteciesNum].y:=y2;
                    Vertecies[VerteciesNum].z:=z2+BlockSize;
                    Move(Tmp2, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2+BlockSize;
                    Vertecies[VerteciesNum].y:=y2;
                    Vertecies[VerteciesNum].z:=z2+BlockSize;
                    Move(Tmp2, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2+BlockSize;
                    Vertecies[VerteciesNum].y:=y2;
                    Vertecies[VerteciesNum].z:=z2;
                    Move(Tmp2, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2;
                    Vertecies[VerteciesNum].y:=y2-BlockSize;
                    Vertecies[VerteciesNum].z:=z2;
                    Move(Tmp, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2;
                    Vertecies[VerteciesNum].y:=y2-BlockSize;
                    Vertecies[VerteciesNum].z:=z2+BlockSize;
                    Move(Tmp, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2+BlockSize;
                    Vertecies[VerteciesNum].y:=y2-BlockSize;
                    Vertecies[VerteciesNum].z:=z2+BlockSize;
                    Move(Tmp, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                    Vertecies[VerteciesNum].x:=x2+BlockSize;
                    Vertecies[VerteciesNum].y:=y2-BlockSize;
                    Vertecies[VerteciesNum].z:=z2;
                    Move(Tmp, Vertecies[VerteciesNum].r, 12);
                    Inc(VerteciesNum);

                 End;
                 Dec(y2, BlockSize);

             End;
             Dec(z2, BlockSize);

         End;
         Inc(x2, BlockSize);

     End;

     // Vertecies transformieren und translieren
     For i:=0 To VerteciesNum-1 Do Begin

         // Drehung um Y-Achse
         a:=(Vertecies[i].x*cy)+(Vertecies[i].z*sy);
         b:=Vertecies[i].y;
         c:=-(Vertecies[i].x*sy)+(Vertecies[i].z*cy);

         // Drehung um X-Achse
         a2:=a;
         b2:=(b*cx)-(c*sx);
         c2:=(b*sx)+(c*cx);

         // Object-Space -> Screen-Space
         Vertecies[i].z:=c2+iZoom; // Nach hinten verschieben
         Vertecies[i].x:=Round((EyeDistance * a2) / Vertecies[i].z)+PreviewBlockCenterX;
         Vertecies[i].y:=Round((EyeDistance * b2) / Vertecies[i].z)+PreviewBlockCenterY;

         // 1 / Z fr W-Buffer...
         Vertecies[i].oow:=1 / Vertecies[i].z;

     End;

     // Blcker des Vorschau-Steins zeichnen
     GrAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
     guColorCombineFunction(GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB);
     grTexClampMode(GR_TMU0, GR_TEXTURECLAMP_WRAP, GR_TEXTURECLAMP_WRAP);
     grTexCombine(GR_TMU0, GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE,
                  GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE, FXFALSE, FXFALSE);
     i:=0;
     BlockTexIndex:=0;
     While i < VerteciesNum Do Begin

           // Textur whlen
           grTexSource(GR_TMU0, aTextureAddresses[TexIndecies[BlockTexIndex]],
                       GR_MIPMAPLEVELMASK_BOTH, aTextures[TexIndecies[BlockTexIndex]]);

           // Unten
           Vertecies[i].tmuvtx[0].sow:=0;
           Vertecies[i].tmuvtx[0].tow:=0;
           Vertecies[i+1].tmuvtx[0].sow:=0;
           Vertecies[i+1].tmuvtx[0].tow:=255 / Vertecies[i+1].z;
           Vertecies[i+2].tmuvtx[0].sow:=255 / Vertecies[i+2].z;
           Vertecies[i+2].tmuvtx[0].tow:=Vertecies[i+2].tmuvtx[0].sow;
           Vertecies[i+3].tmuvtx[0].sow:=255 / Vertecies[i+3].z;
           Vertecies[i+3].tmuvtx[0].tow:=0;
           grDrawTriangle(Vertecies[i], Vertecies[i+1], Vertecies[i+2]);
           grDrawTriangle(Vertecies[i+2], Vertecies[i+3], Vertecies[i]);

           // Oben
           Vertecies[i+4].tmuvtx[0].sow:=0;
           Vertecies[i+4].tmuvtx[0].tow:=255 / Vertecies[i+4].z;
           Vertecies[i+5].tmuvtx[0].sow:=0;
           Vertecies[i+5].tmuvtx[0].tow:=0;
           Vertecies[i+6].tmuvtx[0].sow:=255 / Vertecies[i+6].z;
           Vertecies[i+6].tmuvtx[0].tow:=0;
           Vertecies[i+7].tmuvtx[0].sow:=255 / Vertecies[i+7].z;
           Vertecies[i+7].tmuvtx[0].tow:=Vertecies[i+7].tmuvtx[0].sow;
           grDrawTriangle(Vertecies[i+4], Vertecies[i+7], Vertecies[i+6]);
           grDrawTriangle(Vertecies[i+6], Vertecies[i+5], Vertecies[i+4]);

           // Vorne
           Vertecies[i].tmuvtx[0].sow:=0;
           Vertecies[i].tmuvtx[0].tow:=255 / Vertecies[i+4].z;
           Vertecies[i+3].tmuvtx[0].sow:=255 / Vertecies[i+3].z;
           Vertecies[i+3].tmuvtx[0].tow:=Vertecies[i+3].tmuvtx[0].sow;
           Vertecies[i+4].tmuvtx[0].sow:=0;
           Vertecies[i+4].tmuvtx[0].tow:=0;
           Vertecies[i+7].tmuvtx[0].sow:=255 / Vertecies[i+7].z;
           Vertecies[i+7].tmuvtx[0].tow:=0;
           grDrawTriangle(Vertecies[i+4], Vertecies[i], Vertecies[i+3]);
           grDrawTriangle(Vertecies[i+3], Vertecies[i+7], Vertecies[i+4]);

           // Links
           Vertecies[i].tmuvtx[0].sow:=255 / Vertecies[i].z;
           Vertecies[i].tmuvtx[0].tow:=Vertecies[i].tmuvtx[0].sow;
           Vertecies[i+1].tmuvtx[0].sow:=0;
           Vertecies[i+1].tmuvtx[0].tow:=255 / Vertecies[i+1].z;
           Vertecies[i+4].tmuvtx[0].sow:=255 / Vertecies[i+4].z;
           Vertecies[i+4].tmuvtx[0].tow:=0;
           Vertecies[i+5].tmuvtx[0].sow:=0;
           Vertecies[i+5].tmuvtx[0].tow:=0;
           grDrawTriangle(Vertecies[i+5], Vertecies[i+1], Vertecies[i]);
           grDrawTriangle(Vertecies[i+5], Vertecies[i], Vertecies[i+4]);

           // Hinten
           Vertecies[i+1].tmuvtx[0].sow:=255 / Vertecies[i+1].z;
           Vertecies[i+1].tmuvtx[0].tow:=Vertecies[i+1].tmuvtx[0].sow;
           Vertecies[i+2].tmuvtx[0].sow:=0;
           Vertecies[i+2].tmuvtx[0].tow:=255 / Vertecies[i+2].z;
           Vertecies[i+5].tmuvtx[0].sow:=255 / Vertecies[i+5].z;
           Vertecies[i+5].tmuvtx[0].tow:=0;
           Vertecies[i+6].tmuvtx[0].sow:=0;
           Vertecies[i+6].tmuvtx[0].tow:=0;
           grDrawTriangle(Vertecies[i+6], Vertecies[i+2], Vertecies[i+1]);
           grDrawTriangle(Vertecies[i+6], Vertecies[i+1], Vertecies[i+5]);

           // Rechts
           Vertecies[i+2].tmuvtx[0].sow:=255 / Vertecies[i+2].z;
           Vertecies[i+2].tmuvtx[0].tow:=Vertecies[i+2].tmuvtx[0].sow;
           Vertecies[i+3].tmuvtx[0].sow:=0;
           Vertecies[i+3].tmuvtx[0].tow:=255 / Vertecies[i+3].z;
           Vertecies[i+6].tmuvtx[0].sow:=255 / Vertecies[i+6].z;
           Vertecies[i+6].tmuvtx[0].tow:=0;
           Vertecies[i+7].tmuvtx[0].sow:=0;
           Vertecies[i+7].tmuvtx[0].tow:=0;
           grDrawTriangle(Vertecies[i+7], Vertecies[i+3], Vertecies[i+2]);
           grDrawTriangle(Vertecies[i+7], Vertecies[i+2], Vertecies[i+6]);

           // Nchster Wrfel
           Inc(i, 8);
           Inc(BlockTexIndex);
     End;

     // Weiterdrehen
     sgRotYp:=sgRotYp+(IncFacPerMS.RotY*iFrameTime)/1.5;
     If sgRotXp >= Pi*2 Then sgRotXp:=sgRotXp-(Pi*2);
End;

Procedure TFrmTetris.SetEntry;
Begin
     Palette[Entry]:=LFB_Encode565(r Shr 1, g, b Shr 1);
End;

Procedure TFrmTetris.SetPalette;
Var
   Counter: Longint;
Begin
     For Counter:=0 To 63 Do Begin
         SetEntry(Counter      , 0, 0, Counter*Counter*Counter Div 4096);
         SetEntry(Counter +  64, 0, 0, 63 - (Counter*Counter*Counter Div 4096));
         SetEntry(Counter + 128, Counter*Counter*Counter Div 4096, 0, 0);
         SetEntry(Counter + 192, 63 - (Counter*Counter*Counter Div 4096), 0, 0);
     End;
End;

Procedure TFrmTetris.InitStarsMainMenu;
Var
   i: Integer;
Begin
     For i:=0 To MaxStarsMainMenu Do With aStarsMainMenu[i] Do Begin
         x:=Random(638);
         y:=Random(478);
         Plane:=Random(31);
         Size:=Round(Random(75)/25)+1;
         xInc:=Plane Shr 2;
         If xInc <= 0 Then xInc:=1;
     End;
End;

Procedure TFrmTetris.DrawMainMenu;
Var
   Hor: Array [0..319] Of Integer;
   Ver: Array [0..219] Of Integer;
   Vertecies: Array[0..3] Of TgrVertex;
   MenuVert: Array[0..MaxMenuVertecies] Of TgrVertex;
   Pos: Integer;
   p, t, Radians: Single;
   TimeBeforeDrawing: Cardinal;
   i: Integer;
   r, g, b: Byte;
   Alpha: Integer;
   Col: Word;
   sx, sy, sz, cx, cy, cz: Double;
   a1, b1, c1, a2, b2, c2, a3, b3, c3: Double;
{   Ok: Boolean;
   TmpFace: TFace;}
Begin
     // Timer starten
     timeBeginPeriod(1);
     TimeBeforeDrawing:=timeGetTime;

     // Nur wegen den W-Werten. Seufz.
     grBufferClear(0, 0, GR_WDEPTHVALUE_FARTHEST);

     // Werte fr die Tabellen berechnen
     Radians:=Phase / 20.0;
     For Pos := 0 To 319 Do Begin
         p:=Pos;
         t:=35.0 * sin (cos (Radians / 2) * p / 70);
         t:=t - 50.0 * cos (Radians + (640 - p) / 40);
         Hor[Pos]:=Trunc(t) + 240;
     End;
     Radians:=Phase / 20.0;
     For Pos:=0 To 219 Do Begin
         p:=Pos;
         t:=40.0 * cos (sin (Radians) * p / 25.0);
         t:=t + 70.0 * sin (Radians + (480.0 - p) / 30.0);
         Ver[Pos]:=Trunc(t) + 512;
     End;

     // In PLASMA Zeichnen
     Asm
        pushad

        // Anfangs-Adresse laden
        mov    edi, DWord Ptr Plasma.Adr
        mov    edx, 220

     @YLoop:
        mov    ecx, 320

     @XLoop:
        // Wert fr Horizontal auslesen
        mov    ebx, edx
        sub    ebx, DWord Ptr Hor[ecx*4-4]

        // Wert fr Vertikal auslesen
        mov    eax, ecx
        sub    eax, DWord Ptr Ver[edx*4-4]

        // Die beiden Werte addieren und Farbwert auslesen
        add    ebx, eax
        and    ebx, 255 // Alles was grer ist als 255 abschneiden
        mov    ax, Word Ptr Palette[ebx*2]

        // Pixel in PLASMA schreiben (2x)
        mov    [edi], ax
        mov    [edi+2], ax
        add    edi, 4
        dec    ecx
        jne    @XLoop

        // Fertige Zeile duplizieren
        mov    esi, edi
        sub    esi, 1280
        mov    ecx, 80

     @CopyLoop:
        mov    eax, [esi]
        mov    [edi], eax
        mov    eax, [esi+4]
        mov    [edi+4], eax
        mov    eax, [esi+8]
        mov    [edi+8], eax
        mov    eax, [esi+12]
        mov    [edi+12], eax
        add    esi, 16
        add    edi, 16
        dec    ecx
        jne    @CopyLoop

        dec    edx
        jne    @YLoop

        popad
     End;

     For i:=0 To MaxStarsMainMenu Do With aStarsMainMenu[i] Do Begin
         LFB_Decode565(LFB_GetPixelSprite(x, y, Plasma), r, g, b);
         If b > 0 Then Begin
            r:=(31-b) * Plane Div 31;
            g:=(63-(b Shl 1)) * (Plane Shl 1) Div 63;
         End Else If r > 0 Then Begin
            g:=(63-(r Shl 1)) * (Plane Shl 1) Div 63;
            b:=(32-r) * Plane Div 31;
         End Else Begin
            r:=Plane;
            g:=Plane Shl 1;
            b:=Plane;
         End;
         Col:=LFB_Encode565(r, g, b);
         Case Size Of
              1: LFB_PutPixelSprite(x, y, Col, Plasma);
              2: Begin
                      LFB_PutPixelSprite(x, y, Col, Plasma);
                      LFB_PutPixelSprite(x, y+1, Col, Plasma);
                 End;
              3: Begin
                      LFB_PutPixelSprite(x, y, Col, Plasma);
                      LFB_PutPixelSprite(x+1, y, Col, Plasma);
                      LFB_PutPixelSprite(x, y+1, Col, Plasma);
                 End;
              4: Begin
                      LFB_PutPixelSprite(x, y, Col, Plasma);
                      LFB_PutPixelSprite(x+1, y, Col, Plasma);
                      LFB_PutPixelSprite(x, y+1, Col, Plasma);
                      LFB_PutPixelSprite(x+1, y+1, Col, Plasma);
                 End;
         End;
         Dec(x, xInc);
         If x < 0 Then Begin
            x:=638;
            y:=Random(478);
            Plane:=Random(31);
            Size:=Round(Random(75)/25)+1;
            xInc:=Plane Shr 1;
            If xInc <= 0 Then xInc:=1;
         End;
     End;

     // Plasma zeichnen
     LFB_PutSpriteFast(0, 0, Plasma);

     // Balken am unteren Rand zeichnen
     Vertecies[0].x:=0;
     Vertecies[0].y:=430;
     Vertecies[0].a:=0;
     Vertecies[1].x:=639;
     Vertecies[1].y:=430;
     Vertecies[1].a:=0;
     Vertecies[2].x:=639;
     Vertecies[2].y:=440;
     Vertecies[2].a:=255;
     Vertecies[3].x:=0;
     Vertecies[3].y:=440;
     Vertecies[3].a:=255;
     For i:=0 To 3 Do Begin
         Vertecies[i].r:=0;
         Vertecies[i].g:=0;
         Vertecies[i].b:=0;
         Vertecies[i].oow:=0.5
     End;
     LFB_PutSpriteFast(0, 440, MenuBar);
     grCullMode(GR_CULL_DISABLE);
     guColorCombineFunction(GR_COLORCOMBINE_ITRGB);
     grAlphaCombine(GR_COMBINE_FUNCTION_BLEND_LOCAL, GR_COMBINE_FACTOR_ZERO, GR_COMBINE_LOCAL_ITERATED, GR_COMBINE_OTHER_NONE, FXFALSE);
     grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
     grDrawTriangle(Vertecies[0], Vertecies[1], Vertecies[2]);
     grDrawTriangle(Vertecies[2], Vertecies[3], Vertecies[0]);

     // Sinus- und Cosinus-Werte berechnen (fr Rotation)
     sx:=sin(sgRotXm);
     sy:=sin(sgRotYm);
     sz:=sin(sgRotZm);
     cx:=cos(sgRotXm);
     cy:=cos(sgRotYm);
     cz:=cos(sgRotZm);

     // Menpunkte zeichnen
     For i:=0 To MaxMenuVertecies Do Begin
         MenuVert[i]:=AMenuVertecies[i];

         // Drehung um X-Achse
         a1:=MenuVert[i].x;
         b1:=(MenuVert[i].y*cx)-(MenuVert[i].z*sx);
         c1:=(MenuVert[i].y*sx)+(MenuVert[i].z*cx);

         // Drehung um Y-Achse
         a2:=(a1*cy)+(c1*sy);
         b2:=b1;
         c2:=-(a1*sy)+(c1*cy);

         // Drehung um Z-Achse
         a3:=(a2*cz)-(b2*sz);
         b3:=(a2*sz)+(b2*cz);
         c3:=c2;

         c3:=c3+350;
         MenuVert[i].x:=Round(EyeDistance*a3/c3)+320;
         MenuVert[i].y:=Round(EyeDistance*b3/c3)+220;
         MenuVert[i].a:=50;
         MenuVert[i].oow:=1/c3;
         MenuVert[i].tmuvtx[0].sow:=MenuVert[i].tmuvtx[0].sow / c3;
         MenuVert[i].tmuvtx[0].tow:=MenuVert[i].tmuvtx[0].tow / c3;
         MenuVert[i].tmuvtx[1]:=MenuVert[i].tmuvtx[0];
         MenuVert[i].tmuvtx[1]:=MenuVert[i].tmuvtx[0];
         Alpha:=Round((1/c3*10000 - 30) * 20);
         If Alpha < 0 Then Alpha:=0;
         MenuVert[i].a:=Alpha;
         MenuVert[i].r:=Alpha;
         MenuVert[i].g:=Alpha;
     End;
//     GrAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
     grAlphaBlendFunction(GR_BLEND_SRC_ALPHA, GR_BLEND_ONE_MINUS_SRC_ALPHA, GR_BLEND_ONE, GR_BLEND_ZERO);
     guColorCombineFunction(GR_COLORCOMBINE_TEXTURE_TIMES_ITRGB);
     grTexClampMode(GR_TMU0, GR_TEXTURECLAMP_WRAP, GR_TEXTURECLAMP_WRAP);
     grTexClampMode(GR_TMU1, GR_TEXTURECLAMP_WRAP, GR_TEXTURECLAMP_WRAP);

{     grTexCombine(GR_TMU0, GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE,
                  GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE, FXFALSE, FXFALSE);

     grTexCombine(GR_TMU1, GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL, GR_COMBINE_FACTOR_NONE,
                  GR_COMBINE_FUNCTION_LOCAL, GR_COMBINE_FACTOR_NONE, FXFALSE, FXFALSE);}

     guTexCombineFunction(GR_TMU0, GR_TEXTURECOMBINE_DECAL);
     guTexCombineFunction(GR_TMU1, GR_TEXTURECOMBINE_DECAL);

     grTexSource(GR_TMU0, aTextureAddresses[9], GR_MIPMAPLEVELMASK_BOTH, aTextures[9]);
     grTexSource(GR_TMU1, aTextureAddressesBack[0], GR_MIPMAPLEVELMASK_BOTH, aTexturesBack[0]);
{     Repeat
           Ok:=True;
           For i:=0 To MaxMenuFaces-1 Do Begin
               If ((MenuVert[aMenuFaces[i].a].z+MenuVert[aMenuFaces[i].b].z+MenuVert[aMenuFaces[i].c].z)/3) <
                  ((MenuVert[aMenuFaces[i+1].a].z+MenuVert[aMenuFaces[i+1].b].z+MenuVert[aMenuFaces[i+1].c].z)/3) Then Begin
                  Ok:=False;
                  TmpFace:=aMenuFaces[i+1];
                  aMenuFaces[i+1]:=aMenuFaces[i];
                  aMenuFaces[i]:=TmpFace;
               End;
           End;
     Until Ok;}
     For i:=0 To MaxMenuFaces Do Begin
         grDrawTriangle(MenuVert[aMenuFaces[i].a],
                        MenuVert[aMenuFaces[i].b],
                        MenuVert[aMenuFaces[i].c]);
     End;
     grCullMode(GR_CULL_POSITIVE);

     // Sprites zeichnen
     If LFB_Lock <> Nil Then
//        LFB_PutSprite(219, 0, MenuLogo);
        OutTextXY(0, 0, 0, 5, Stri(Round(fps)));
     LFB_UnLock;

     // Berechneten Frame anzeigen
     grBufferSwap(1);

     // Frame-Zeit stoppen
     iFrameTime:=timeGetTime-TimeBeforeDrawing;
     timeEndPeriod(1);
     // Fr den unwahrscheinlichen Fall, da der Rechner mehr als 1000 f/s schafft...
     If iFrameTime <= 0 Then iFrameTime:=1;
     fps:=1000/iFrameTime;

     // Damit sich das Plasma schn bewegt...
     Inc(Phase);

     sgRotXm:=sgRotXm - (Pi / ((MenuItems / 2) * 10)) * iRotMenu;
     If sgRotXm < 0 Then sgRotXm:=2 * Pi + sgRotXm;
     If sgRotXm > 2 * Pi Then sgRotXm:=sgRotXm - 2 * Pi;
     If iRotMenu <> 0 Then Begin
        For i:=0 To MenuItems Do Begin
            If Abs(sgRotXm-(i * (pi / (MenuItems / 2)))) <= 0.06 Then iRotMenu:=0;
        End;
     End;
End;

Procedure TFrmTetris.InitMenuItems;
Var
   i, m: Integer;
   Angle: Single;
   AngleInc: Single;
Begin
     AngleInc:=((2*Pi)/(MenuItems*(5/4)));
     i:=0;
     For m:=0 To MenuItems-1 Do Begin
         Angle:=m*(2*Pi/MenuItems);
         aMenuVertecies[i].x:=-128;
         aMenuVertecies[i].y:=Round((MenuItemRadius*Cos(Angle))-(MenuItemRadius*Sin(Angle)));
         aMenuVertecies[i].z:=Round((MenuItemRadius*Sin(Angle))+(MenuItemRadius*Cos(Angle)));
         aMenuVertecies[i].tmuvtx[0].sow:=0;
         aMenuVertecies[i].tmuvtx[0].tow:=0;
         aMenuVertecies[i+1].x:=128;
         aMenuVertecies[i+1].y:=aMenuVertecies[i].y;
         aMenuVertecies[i+1].z:=aMenuVertecies[i].z;
         aMenuVertecies[i+1].tmuvtx[0].sow:=255;
         aMenuVertecies[i+1].tmuvtx[0].tow:=0;
         Angle:=Angle+AngleInc;
         aMenuVertecies[i+2].x:=128;
         aMenuVertecies[i+2].y:=Round((MenuItemRadius*Cos(Angle))-(MenuItemRadius*Sin(Angle)));
         aMenuVertecies[i+2].z:=Round((MenuItemRadius*Sin(Angle))+(MenuItemRadius*Cos(Angle)));
         aMenuVertecies[i+2].tmuvtx[0].sow:=255;
         aMenuVertecies[i+2].tmuvtx[0].tow:=63;
         aMenuVertecies[i+3].x:=-128;
         aMenuVertecies[i+3].y:=aMenuVertecies[i+2].y;
         aMenuVertecies[i+3].z:=aMenuVertecies[i+2].z;
         aMenuVertecies[i+3].tmuvtx[0].sow:=0;
         aMenuVertecies[i+3].tmuvtx[0].tow:=63;
         Inc(i, 4);
     End;
     For i:=0 To MaxMenuVertecies Do Begin
         aMenuVertecies[i].r:=255;
         aMenuVertecies[i].g:=255;
         aMenuVertecies[i].b:=255;
     End;
     i:=0;
     For m:=0 To MenuItems-1 Do Begin
         aMenuFaces[i].a:=(m Shl 2);
         aMenuFaces[i].b:=(m Shl 2)+1;
         aMenuFaces[i].c:=(m Shl 2)+2;
         aMenuFaces[i+1].a:=(m Shl 2)+2;
         aMenuFaces[i+1].b:=(m Shl 2)+3;
         aMenuFaces[i+1].c:=(m Shl 2);
         Inc(i, 2);
     End;
End;

End.
